/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  Pseudo_QC_Float.qpp

  Qore Programming Language

  Copyright (C) 2003 - 2023 Qore Technologies, s.r.o.

  Permission is hereby granted, free of charge, to any person obtaining a
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation
  the rights to use, copy, modify, merge, publish, distribute, sublicense,
  and/or sell copies of the Software, and to permit persons to whom the
  Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.

  Note that the Qore library is released under a choice of three open-source
  licenses: MIT (as above), LGPL 2+, or GPL 2+; see README-LICENSE for more
  information.
*/

#include <qore/Qore.h>
#include "qore/intern/qore_number_private.h"

#include <cmath>
/* Qore class Qore::<float> */

qore_classid_t CID_PSEUDOFLOAT;
QoreClass* QC_PSEUDOFLOAT;

// float PseudoFloat::abs(){}
static QoreValue PseudoFloat_abs(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    double f = v.getAsFloat();
# 221 "Pseudo_QC_Float.qpp"
    return fabs(f);
}

// string PseudoFloat::format(string fmt){}
static QoreValue PseudoFloat_format_Vs(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* fmt = HARD_QORE_VALUE_STRING(args, 0);
    double f = v.getAsFloat();
# 131 "Pseudo_QC_Float.qpp"
    return format_float_intern(*fmt, f, xsink);
}

// string PseudoFloat::format(int prec = 2, string decimal_sep = ".", string thousands_sep = ""){}
static QoreValue PseudoFloat_format_ViVsVs(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 prec = HARD_QORE_VALUE_INT(args, 0);
    const QoreStringNode* decimal_sep = HARD_QORE_VALUE_STRING(args, 1);
    const QoreStringNode* thousands_sep = HARD_QORE_VALUE_STRING(args, 2);
    double f = v.getAsFloat();
# 157 "Pseudo_QC_Float.qpp"
    return format_float_intern(prec, decimal_sep, thousands_sep, f, xsink);
}

// bool PseudoFloat::infp(){}
static QoreValue PseudoFloat_infp(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    double f = v.getAsFloat();
# 189 "Pseudo_QC_Float.qpp"
    return std::isinf(f) != 0;
}

// bool PseudoFloat::intp(){}
static QoreValue PseudoFloat_intp(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 87 "Pseudo_QC_Float.qpp"
    return true;
}

// bool PseudoFloat::nanp(){}
static QoreValue PseudoFloat_nanp(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    double f = v.getAsFloat();
# 173 "Pseudo_QC_Float.qpp"
    return std::isnan(f) != 0;
}

// int PseudoFloat::sign(){}
static QoreValue PseudoFloat_sign(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    double f = v.getAsFloat();
# 203 "Pseudo_QC_Float.qpp"
    if (!f)
      return 0;
   return f > 0.0 ? 1 : -1;
}

// bool PseudoFloat::strp(){}
static QoreValue PseudoFloat_strp(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 101 "Pseudo_QC_Float.qpp"
    return true;
}

// string PseudoFloat::toBase(int base = 10){}
static QoreValue PseudoFloat_toBase_Vi(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 base = HARD_QORE_VALUE_INT(args, 0);
    double f = v.getAsFloat();
# 244 "Pseudo_QC_Float.qpp"
    return qore_number_private::toBase(f, base, xsink);
}

// int PseudoFloat::typeCode(){}
static QoreValue PseudoFloat_typeCode(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 56 "Pseudo_QC_Float.qpp"
    return NT_FLOAT;
}

// bool PseudoFloat::val(){}
static QoreValue PseudoFloat_val(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    double f = v.getAsFloat();
# 73 "Pseudo_QC_Float.qpp"
    return (bool)f == 0.0 ? false : true;
}

DLLLOCAL void preinitPseudoFloatClass() {
    QC_PSEUDOFLOAT = new QoreClass("<float>", "::Qore::<float>", QDOM_DEFAULT);
    CID_PSEUDOFLOAT = QC_PSEUDOFLOAT->getID();
    QC_PSEUDOFLOAT->setSystem();
}

DLLLOCAL QoreClass* initPseudoFloatClass() {
    if (!QC_PSEUDOFLOAT)
        preinitPseudoFloatClass();

    // set parent class
    assert(QC_PSEUDOVALUE);
    QC_PSEUDOFLOAT->addBuiltinVirtualBaseClass(QC_PSEUDOVALUE);

    // float PseudoFloat::abs(){}
    QC_PSEUDOFLOAT->addMethod("abs", (q_method_n_t)PseudoFloat_abs, Public, QCF_CONSTANT, QDOM_DEFAULT, floatTypeInfo);

    // string PseudoFloat::format(string fmt){}
    QC_PSEUDOFLOAT->addMethod("format", (q_method_n_t)PseudoFloat_format_Vs, Public, QCF_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "fmt");

    // string PseudoFloat::format(int prec = 2, string decimal_sep = ".", string thousands_sep = ""){}
    QC_PSEUDOFLOAT->addMethod("format", (q_method_n_t)PseudoFloat_format_ViVsVs, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 3, bigIntTypeInfo, QoreSimpleValue().assign((int64)2), "prec", stringTypeInfo, QoreSimpleValue().assign(new QoreStringNode(".")), "decimal_sep", stringTypeInfo, QoreSimpleValue().assign(new QoreStringNode("")), "thousands_sep");

    // bool PseudoFloat::infp(){}
    QC_PSEUDOFLOAT->addMethod("infp", (q_method_n_t)PseudoFloat_infp, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    // bool PseudoFloat::intp(){}
    QC_PSEUDOFLOAT->addMethod("intp", (q_method_n_t)PseudoFloat_intp, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    // bool PseudoFloat::nanp(){}
    QC_PSEUDOFLOAT->addMethod("nanp", (q_method_n_t)PseudoFloat_nanp, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    // int PseudoFloat::sign(){}
    QC_PSEUDOFLOAT->addMethod("sign", (q_method_n_t)PseudoFloat_sign, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // bool PseudoFloat::strp(){}
    QC_PSEUDOFLOAT->addMethod("strp", (q_method_n_t)PseudoFloat_strp, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    // string PseudoFloat::toBase(int base = 10){}
    QC_PSEUDOFLOAT->addMethod("toBase", (q_method_n_t)PseudoFloat_toBase_Vi, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, bigIntTypeInfo, QoreSimpleValue().assign((int64)10), "base");

    // int PseudoFloat::typeCode(){}
    QC_PSEUDOFLOAT->addMethod("typeCode", (q_method_n_t)PseudoFloat_typeCode, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // bool PseudoFloat::val(){}
    QC_PSEUDOFLOAT->addMethod("val", (q_method_n_t)PseudoFloat_val, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    return QC_PSEUDOFLOAT;
}
