/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
  Pseudo_QC_Int.qpp

  Qore Programming Language

  Copyright (C) 2003 - 2023 Qore Technologies, s.r.o.

  Permission is hereby granted, free of charge, to any person obtaining a
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation
  the rights to use, copy, modify, merge, publish, distribute, sublicense,
  and/or sell copies of the Software, and to permit persons to whom the
  Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.

  Note that the Qore library is released under a choice of three open-source
  licenses: MIT (as above), LGPL 2+, or GPL 2+; see README-LICENSE for more
  information.
*/

#include <qore/Qore.h>
#include "qore/intern/qore_number_private.h"

string to_base_intern(int64 i, int base);
/* Qore class Qore::<int> */

qore_classid_t CID_PSEUDOINT;
QoreClass* QC_PSEUDOINT;

// int PseudoInt::abs(){}
static QoreValue PseudoInt_abs(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 i = v.getAsBigInt();
# 303 "Pseudo_QC_Int.qpp"
    return i < 0 ? -i : i;
}

// binary PseudoInt::encodeLsb(int size = 4){}
static QoreValue PseudoInt_encodeLsb_Vi(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 size = HARD_QORE_VALUE_INT(args, 0);
    int64 i = v.getAsBigInt();
# 250 "Pseudo_QC_Int.qpp"
    SimpleRefHolder<BinaryNode> b(new BinaryNode);
   switch (size) {
      case 1: {
         unsigned char c = i;
         if (i != c)
            return xsink->raiseException("SIZE-ERROR", "the integer value called with <int>::encodeLsb(size=1) (" QLLD ") is too large to fit in 1 byte", i);
         b->append((void*)&c, 1);
         break;
      }

      case 2: {
         uint16_t num = i;
         if (i != num)
            return xsink->raiseException("SIZE-ERROR", "the integer value called with <int>::encodeLsb(size=2) (" QLLD ") is too large to fit in 2 bytes", i);
         num = i2LSB(num);
         b->append((void*)&num, 2);
         break;
      }

      case 4: {
         uint32_t num = i;
         if (i != num)
            return xsink->raiseException("SIZE-ERROR", "the integer value called with <int>::encodeLsb(size=4) (" QLLD ") is too large to fit in 4 bytes", i);
         num = i4LSB(num);
         b->append((void*)&num, 4);
         break;
      }

      case 8: {
         int64 num = i8LSB(i);
         b->append((void*)&num, 8);
         break;
      }

      default:
         return xsink->raiseException("SIZE-ERROR", "the size given to <int>::encodeLsb(%d) is not 1, 2, 4, or 8", size);
   }
   return b.release();
}

// binary PseudoInt::encodeMsb(int size = 4){}
static QoreValue PseudoInt_encodeMsb_Vi(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 size = HARD_QORE_VALUE_INT(args, 0);
    int64 i = v.getAsBigInt();
# 195 "Pseudo_QC_Int.qpp"
    SimpleRefHolder<BinaryNode> b(new BinaryNode);
   switch (size) {
      case 1: {
         unsigned char c = i;
         if (i != c)
            return xsink->raiseException("SIZE-ERROR", "the integer value called with <int>::encodeMsb(size=1) (" QLLD ") is too large to fit in 1 byte", i);
         b->append((void*)&c, 1);
         break;
      }

      case 2: {
         uint16_t num = i;
         if (i != num)
            return xsink->raiseException("SIZE-ERROR", "the integer value called with <int>::encodeMsb(size=2) (" QLLD ") is too large to fit in 2 bytes", i);
         num = htons(num);
         b->append((void*)&num, 2);
         break;
      }

      case 4: {
         uint32_t num = i;
         if (i != num)
            return xsink->raiseException("SIZE-ERROR", "the integer value called with <int>::encodeMsb(size=4) (" QLLD ") is too large to fit in 4 bytes", i);
         num = htonl(num);
         b->append((void*)&num, 4);
         break;
      }

      case 8: {
         int64 num = i8MSB(i);
         b->append((void*)&num, 8);
         break;
      }

      default:
         return xsink->raiseException("SIZE-ERROR", "the size given to <int>::encodeMsb(" QLLD ") is not 1, 2, 4, or 8", size);
   }
   return b.release();
}

// string PseudoInt::format(string fmt){}
static QoreValue PseudoInt_format_Vs(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* fmt = HARD_QORE_VALUE_STRING(args, 0);
    int64 i = v.getAsBigInt();
# 129 "Pseudo_QC_Int.qpp"
    SimpleRefHolder<QoreStringNode> rv(new QoreStringNode(fmt->getEncoding()));
   rv->sprintf(QLLD, i);
   return qore_number_private::formatNumberString(**rv, *fmt, xsink) ? 0 : rv.release();
}

// string PseudoInt::format(int prec = 2, string decimal_sep = ".", string thousands_sep = ""){}
static QoreValue PseudoInt_format_ViVsVs(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 prec = HARD_QORE_VALUE_INT(args, 0);
    const QoreStringNode* decimal_sep = HARD_QORE_VALUE_STRING(args, 1);
    const QoreStringNode* thousands_sep = HARD_QORE_VALUE_STRING(args, 2);
    int64 i = v.getAsBigInt();
# 157 "Pseudo_QC_Int.qpp"
    SimpleRefHolder<QoreStringNode> rv(new QoreStringNode);
   rv->sprintf(QLLD, i);
   if (qore_number_private::formatNumberString(**rv, prec, decimal_sep, thousands_sep, xsink))
      return 0;
   return rv.release();
}

// bool PseudoInt::intp(){}
static QoreValue PseudoInt_intp(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 87 "Pseudo_QC_Int.qpp"
    return true;
}

// int PseudoInt::sign(){}
static QoreValue PseudoInt_sign(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 i = v.getAsBigInt();
# 175 "Pseudo_QC_Int.qpp"
    if (!i)
      return 0;
   return i > 0 ? 1 : -1;
}

// bool PseudoInt::strp(){}
static QoreValue PseudoInt_strp(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 101 "Pseudo_QC_Int.qpp"
    return true;
}

// string PseudoInt::toBase(int base = 10){}
static QoreValue PseudoInt_toBase_Vi(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 base = HARD_QORE_VALUE_INT(args, 0);
    int64 i = v.getAsBigInt();
# 358 "Pseudo_QC_Int.qpp"
    if (base < 2 || base > 36)
        return xsink->raiseException("INVALID-BASE", "base " QLLD " is invalid; base must be 2 - 36 inclusive", base);

    string s = to_base_intern(i, base);
    return new QoreStringNode(s);
}

// string PseudoInt::toUnicode(){}
static QoreValue PseudoInt_toUnicode(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 i = v.getAsBigInt();
# 319 "Pseudo_QC_Int.qpp"
    SimpleRefHolder<QoreStringNode> str(new QoreStringNode(QCS_UTF8));
   str->concatUnicode(i);
   return *xsink ? 0 : str.release();
}

// int PseudoInt::typeCode(){}
static QoreValue PseudoInt_typeCode(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 56 "Pseudo_QC_Int.qpp"
    return NT_INT;
}

// bool PseudoInt::val(){}
static QoreValue PseudoInt_val(QoreObject* self, const QoreValue& v, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 i = v.getAsBigInt();
# 73 "Pseudo_QC_Int.qpp"
    return (bool)i;
}

DLLLOCAL void preinitPseudoIntClass() {
    QC_PSEUDOINT = new QoreClass("<int>", "::Qore::<int>", QDOM_DEFAULT);
    CID_PSEUDOINT = QC_PSEUDOINT->getID();
    QC_PSEUDOINT->setSystem();
}

DLLLOCAL QoreClass* initPseudoIntClass() {
    if (!QC_PSEUDOINT)
        preinitPseudoIntClass();

    // set parent class
    assert(QC_PSEUDOVALUE);
    QC_PSEUDOINT->addBuiltinVirtualBaseClass(QC_PSEUDOVALUE);

    // int PseudoInt::abs(){}
    QC_PSEUDOINT->addMethod("abs", (q_method_n_t)PseudoInt_abs, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // binary PseudoInt::encodeLsb(int size = 4){}
    QC_PSEUDOINT->addMethod("encodeLsb", (q_method_n_t)PseudoInt_encodeLsb_Vi, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, bigIntTypeInfo, QoreSimpleValue().assign((int64)4), "size");

    // binary PseudoInt::encodeMsb(int size = 4){}
    QC_PSEUDOINT->addMethod("encodeMsb", (q_method_n_t)PseudoInt_encodeMsb_Vi, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo, 1, bigIntTypeInfo, QoreSimpleValue().assign((int64)4), "size");

    // string PseudoInt::format(string fmt){}
    QC_PSEUDOINT->addMethod("format", (q_method_n_t)PseudoInt_format_Vs, Public, QCF_CONSTANT, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "fmt");

    // string PseudoInt::format(int prec = 2, string decimal_sep = ".", string thousands_sep = ""){}
    QC_PSEUDOINT->addMethod("format", (q_method_n_t)PseudoInt_format_ViVsVs, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 3, bigIntTypeInfo, QoreSimpleValue().assign((int64)2), "prec", stringTypeInfo, QoreSimpleValue().assign(new QoreStringNode(".")), "decimal_sep", stringTypeInfo, QoreSimpleValue().assign(new QoreStringNode("")), "thousands_sep");

    // bool PseudoInt::intp(){}
    QC_PSEUDOINT->addMethod("intp", (q_method_n_t)PseudoInt_intp, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    // int PseudoInt::sign(){}
    QC_PSEUDOINT->addMethod("sign", (q_method_n_t)PseudoInt_sign, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // bool PseudoInt::strp(){}
    QC_PSEUDOINT->addMethod("strp", (q_method_n_t)PseudoInt_strp, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    // string PseudoInt::toBase(int base = 10){}
    QC_PSEUDOINT->addMethod("toBase", (q_method_n_t)PseudoInt_toBase_Vi, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo, 1, bigIntTypeInfo, QoreSimpleValue().assign((int64)10), "base");

    // string PseudoInt::toUnicode(){}
    QC_PSEUDOINT->addMethod("toUnicode", (q_method_n_t)PseudoInt_toUnicode, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo);

    // int PseudoInt::typeCode(){}
    QC_PSEUDOINT->addMethod("typeCode", (q_method_n_t)PseudoInt_typeCode, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // bool PseudoInt::val(){}
    QC_PSEUDOINT->addMethod("val", (q_method_n_t)PseudoInt_val, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    return QC_PSEUDOINT;
}
string to_base_intern(int64 i, int base) {
    if (i < 0)
        return '-' + to_base_intern(-i, base);

    string s;
    while (true) {
        int rem = i % base;
        s = char((rem < 10 ? 48 : 55) + rem) + s;
        if ((i /= base) == 0)
            break;
    }
    return s;
}
