/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
    QC_DatasourcePool.cpp

    Qore Programming Language

    Copyright (C) 2003 - 2023 Qore Technologies, s.r.o.

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

    Note that the Qore library is released under a choice of three open-source
    licenses: MIT (as above), LGPL 2+, or GPL 2+; see README-LICENSE for more
    information.
*/

#include <qore/Qore.h>
#include <qore/DBI.h>

#include "qore/intern/QC_DatasourcePool.h"
#include "qore/intern/QC_Serializable.h"

DLLLOCAL extern QoreClass* QC_ABSTRACTDATASOURCE;
DLLLOCAL extern QoreClass* QC_ABSTRACTSQLSTATEMENT;
DLLLOCAL extern QoreClass* QC_SQLSTATEMENT;

#define DP_MIN 3
#define DP_MAX 10

static const char *DSPC_ERR = "DATASOURCEPOOL-CONSTRUCTOR-ERROR";

static void dsp_constructor_hash(const QoreHashNode* opts, QoreObject* self, Queue* queue, const QoreValue& arg, ExceptionSink* xsink) {
   ReferenceHolder<Queue> q(queue, xsink);

   const char *str = check_hash_key(opts, "type", DSPC_ERR, xsink);
   if (*xsink)
      return;

   if (!str) {
      xsink->raiseException(DSPC_ERR, "expecting a string value with the 'type' key giving the driver name");
      return;
   }

   DBIDriver* db_driver = DBI.find(str);
   if (!db_driver) {
      if (!*xsink)
         xsink->raiseException("DATASOURCEPOOL-UNSUPPORTED-DATABASE", "no DBI driver can be found for database type '%s'", str);
      return;
   }

   const char* user = check_hash_key(opts, "user", DSPC_ERR, xsink);
   if (*xsink) return;

   const char* pass = check_hash_key(opts, "pass", DSPC_ERR, xsink);
   if (*xsink) return;

   const char* db = check_hash_key(opts, "db", DSPC_ERR, xsink);
   if (*xsink) return;

   const char* charset = check_hash_key(opts, "charset", DSPC_ERR, xsink);
   if (*xsink) return;

   const char* host = check_hash_key(opts, "host", DSPC_ERR, xsink);
   if (*xsink) return;

   bool found;
   int port = (int)opts->getKeyValue("port").getAsBigInt();
   if (port < 0) {
      xsink->raiseException(DSPC_ERR, "port value must be zero (meaning use the default port) or positive (value given: %d)", port);
      return;
   }

   // get options
   int min = 0, max = 0;

   QoreValue p = opts->getKeyValue("options");
   const QoreHashNode* opt_hash;
   if (!p.isNothing()) {
      if (p.getType() != NT_HASH) {
         xsink->raiseException(DSPC_ERR, "'options' key is not type 'hash', instead got type '%s'", p.getTypeName());
         return;
      }

      opt_hash = p.get<const QoreHashNode>();

      min = (int)opt_hash->getKeyValueExistence("min", found, xsink).getAsBigInt();
      if (*xsink)
         return;
      if (found) {
         if (min < 0) {
            xsink->raiseException(DSPC_ERR, "minimum connections must be > 0 (value given: %d)", min);
            return;
         }
      }

      max = (int)opt_hash->getKeyValueExistence("max", found, xsink).getAsBigInt();
      if (*xsink)
         return;
      if (found) {
         if (max < min) {
            xsink->raiseException(DSPC_ERR, "maximum connections must be >= min(%d) (value given: %d)", min, max);
            return;
         }
         if (!max) {
            xsink->raiseException(DSPC_ERR, "maximum connections must be > 0 (value given: %d)", min, max);
            return;
         }
      }
   }
   else
      opt_hash = nullptr;

   if (!min)
      min = DP_MIN;

   if (!max)
      max = DP_MAX;

   SimpleRefHolder<DatasourcePool> ds(new DatasourcePool(xsink, db_driver, user, pass, db, charset, host, min, max, port, opt_hash, q.release(), arg.refSelf()));
   if (!*xsink)
      self->setPrivate(CID_DATASOURCEPOOL, ds.release());
   else
      ds->destructor(xsink);
}

static QoreHashNode* DatasourcePool_serializer(const QoreObject& self, const DatasourcePool& dsp, QoreSerializationContext& context, ExceptionSink* xsink) {
    SimpleRefHolder<QoreStringNode> str(const_cast<DatasourcePool&>(dsp).getConfigString(xsink));
    if (*xsink) {
        return nullptr;
    }

    ReferenceHolder<QoreHashNode> rv(new QoreHashNode(autoTypeInfo), xsink);
    rv->setKeyValue("conf", str.release(), xsink);
    return rv.release();
}

static void DatasourcePool_deserializer(QoreObject& self, const QoreHashNode* sdata, QoreDeserializationContext& context, ExceptionSink* xsink) {
    if (sdata) {
        QoreValue val = sdata->getKeyValue("conf");
        if (val) {
            if (val.getType() != NT_STRING) {
                xsink->raiseException("DESERIALIZATION-ERROR", "DatasourcePool serialization key 'conf' has type '%s'; expecting 'string'", val.getTypeName());
                return;
            }

            ReferenceHolder<QoreHashNode> h(parseDatasource(val.get<QoreStringNode>()->c_str(), xsink), xsink);
            if (!h) {
                assert(*xsink);
                return;
            }

            dsp_constructor_hash(*h, &self, nullptr, QoreValue(), xsink);
            return;
        }
    }

    xsink->raiseException("DESERIALIZATION-ERROR", "missing serialization data for the DatasourcePool class");
}
/* Qore class Qore::SQL::DatasourcePool */

qore_classid_t CID_DATASOURCEPOOL;
QoreClass* QC_DATASOURCEPOOL;

// nothing DatasourcePool::beginTransaction(){}
static QoreValue DatasourcePool_beginTransaction(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 639 "QC_DatasourcePool.qpp"
    ds->beginTransaction(xsink);
    return QoreValue();
}

// nothing DatasourcePool::clearEventQueue(){}
static QoreValue DatasourcePool_clearEventQueue(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 1045 "QC_DatasourcePool.qpp"
    ds->setEventQueue(0, 0, xsink);
    return QoreValue();
}

// nothing DatasourcePool::clearWarningCallback(){}
static QoreValue DatasourcePool_clearWarningCallback(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 948 "QC_DatasourcePool.qpp"
    ds->clearWarningCallback(xsink);
    return QoreValue();
}

// nothing DatasourcePool::commit(){}
static QoreValue DatasourcePool_commit(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 371 "QC_DatasourcePool.qpp"
    ds->commit(xsink);
    return QoreValue();
}

// DatasourcePool::constructor(string driver, *string user, *string pass, *string db, *string encoding, *string host, softint min = 3, softint max = 10, softint port = 0, *Qore::Thread::Queue queue, auto arg) {}
static void DatasourcePool_constructor_VsNsNsNsNsNsviviviC5QueueVA(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* driver = HARD_QORE_VALUE_STRING(args, 0);
    const QoreStringNode* user = get_param_value(args, 1).get<const QoreStringNode>();
    const QoreStringNode* pass = get_param_value(args, 2).get<const QoreStringNode>();
    const QoreStringNode* db = get_param_value(args, 3).get<const QoreStringNode>();
    const QoreStringNode* encoding = get_param_value(args, 4).get<const QoreStringNode>();
    const QoreStringNode* host = get_param_value(args, 5).get<const QoreStringNode>();
    int64 min = HARD_QORE_VALUE_INT(args, 6);
    int64 max = HARD_QORE_VALUE_INT(args, 7);
    int64 port = HARD_QORE_VALUE_INT(args, 8);
    HARD_QORE_VALUE_OBJ_OR_NOTHING_DATA(queue, Queue, args, 9, CID_QUEUE, xsink);
    if (*xsink)
        return;
    QoreValue arg = get_param_value(args, 10);
# 242 "QC_DatasourcePool.qpp"
if (!port && min == 3 && max == 10 && driver->find(':') != -1) {
      ReferenceHolder<QoreHashNode> h(parseDatasource(driver->getBuffer(), xsink), xsink);
      if (!h) {
         assert(*xsink);
         return;
      }
      dsp_constructor_hash(*h, self, queue, arg, xsink);
      return;
   }

   ReferenceHolder<Queue> q(queue, xsink);

   DBIDriver *db_driver = DBI.find(driver->getBuffer());
   if (!db_driver) {
      xsink->raiseException("DATASOURCEPOOL-UNSUPPORTED-DATABASE", "no DBI driver can be found for database type '%s'", driver->getBuffer());
      return;
   }

   if (min <= 0) {
      xsink->raiseException(DSPC_ERR, "minimum connections must be > 0 (value given: " QLLD ")", min);
      return;
   }

   if (max < min) {
      xsink->raiseException(DSPC_ERR, "maximum connections must be >= min(%d) (value given: " QLLD ")", min, max);
      return;
   }

   if (port < 0) {
      xsink->raiseException(DSPC_ERR, "port value must be zero (meaning use the default port) or positive (value given: " QLLD ")", port);
      return;
   }

   SimpleRefHolder<DatasourcePool> ds(new DatasourcePool(xsink, db_driver, user ? user->getBuffer() : 0, pass ? pass->getBuffer() : 0,
							 db ? db->getBuffer() : 0, encoding ? encoding->getBuffer() : 0, host ? host->getBuffer() : 0,
							 min, max, port, 0, q.release(), arg.refSelf()));
   if (!*xsink)
      self->setPrivate(CID_DATASOURCEPOOL, ds.release());
   else
      ds->destructor(xsink);
}

// DatasourcePool::constructor(string desc, *Qore::Thread::Queue queue, auto arg) {}
static void DatasourcePool_constructor_VsC5QueueVA(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* desc = HARD_QORE_VALUE_STRING(args, 0);
    HARD_QORE_VALUE_OBJ_OR_NOTHING_DATA(queue, Queue, args, 1, CID_QUEUE, xsink);
    if (*xsink)
        return;
    QoreValue arg = get_param_value(args, 2);
# 301 "QC_DatasourcePool.qpp"
ReferenceHolder<Queue> q(queue, xsink);
    ReferenceHolder<QoreHashNode> h(parseDatasource(desc->getBuffer(), xsink), xsink);
    if (!h) {
        assert(*xsink);
        return;
    }

    dsp_constructor_hash(*h, self, q.release(), arg, xsink);
}

// DatasourcePool::constructor(hash<auto> opts, *Qore::Thread::Queue queue, auto arg) {}
static void DatasourcePool_constructor_C10hash_auto_C5QueueVA(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreHashNode* opts = HARD_QORE_VALUE_HASH(args, 0);
    HARD_QORE_VALUE_OBJ_OR_NOTHING_DATA(queue, Queue, args, 1, CID_QUEUE, xsink);
    if (*xsink)
        return;
    QoreValue arg = get_param_value(args, 2);
# 334 "QC_DatasourcePool.qpp"
dsp_constructor_hash(opts, self, queue, arg, xsink);
}

// DatasourcePool::copy() {}
static void DatasourcePool_copy(QoreObject* self, QoreObject* old, DatasourcePool* ds, ExceptionSink* xsink) {
# 359 "QC_DatasourcePool.qpp"
self->setPrivate(CID_DATASOURCEPOOL, new DatasourcePool(*ds, xsink));
}

// bool DatasourcePool::currentThreadInTransaction(){}
static QoreValue DatasourcePool_currentThreadInTransaction(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 885 "QC_DatasourcePool.qpp"
    return ds->currentThreadInTransaction();
}

// DatasourcePool::destructor() {}
static void DatasourcePool_destructor(QoreObject* self, DatasourcePool* ds, ExceptionSink* xsink) {
# 347 "QC_DatasourcePool.qpp"
ds->destructor(xsink);
   ds->deref();
}

// auto DatasourcePool::exec(string sql, ...){}
static QoreValue DatasourcePool_exec_VsVV(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* sql = HARD_QORE_VALUE_STRING(args, 0);
# 402 "QC_DatasourcePool.qpp"
    ReferenceHolder<QoreListNode> vargs(args->size() > 1 ? args->copyListFrom(1) : 0, xsink);
   return ds->exec(sql, *vargs, xsink);
}

// auto DatasourcePool::execRaw(string sql){}
static QoreValue DatasourcePool_execRaw_Vs(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* sql = HARD_QORE_VALUE_STRING(args, 0);
# 445 "QC_DatasourcePool.qpp"
    return ds->execRaw(sql, xsink);
}

// int DatasourcePool::getCapabilities(){}
static QoreValue DatasourcePool_getCapabilities(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 1061 "QC_DatasourcePool.qpp"
    return ds->getCapabilities();
}

// list<auto> DatasourcePool::getCapabilityList(){}
static QoreValue DatasourcePool_getCapabilityList(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 1077 "QC_DatasourcePool.qpp"
    return ds->getCapabilityList();
}

// auto DatasourcePool::getClientVersion(){}
static QoreValue DatasourcePool_getClientVersion(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 859 "QC_DatasourcePool.qpp"
    return ds->getClientVersion(xsink);
}

// hash<auto> DatasourcePool::getConfigHash(){}
static QoreValue DatasourcePool_getConfigHash(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 922 "QC_DatasourcePool.qpp"
    return ds->getConfigHash(xsink);
}

// string DatasourcePool::getConfigString(){}
static QoreValue DatasourcePool_getConfigString(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 936 "QC_DatasourcePool.qpp"
    return ds->getConfigString(xsink);
}

// *string DatasourcePool::getDBCharset(){}
static QoreValue DatasourcePool_getDBCharset(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 689 "QC_DatasourcePool.qpp"
    return ds->getPendingDBEncoding();
}

// *string DatasourcePool::getDBEncoding(){}
static QoreValue DatasourcePool_getDBEncoding(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 703 "QC_DatasourcePool.qpp"
    return ds->getPendingDBEncoding();
}

// *string DatasourcePool::getDBName(){}
static QoreValue DatasourcePool_getDBName(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 675 "QC_DatasourcePool.qpp"
    return ds->getPendingDBName();
}

// string DatasourcePool::getDriverName(){}
static QoreValue DatasourcePool_getDriverName(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 770 "QC_DatasourcePool.qpp"
    return new QoreStringNode(ds->getDriverName());
}

// string DatasourcePool::getDriverRealName(){}
static QoreValue DatasourcePool_getDriverRealName(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 791 "QC_DatasourcePool.qpp"
    return ds->getDriverRealName(xsink);
}

// int DatasourcePool::getErrorTimeout(){}
static QoreValue DatasourcePool_getErrorTimeout(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 1019 "QC_DatasourcePool.qpp"
    return ds->getErrorTimeout();
}

// *string DatasourcePool::getHostName(){}
static QoreValue DatasourcePool_getHostName(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 743 "QC_DatasourcePool.qpp"
    return ds->getPendingHostName();
}

// int DatasourcePool::getMaximum(){}
static QoreValue DatasourcePool_getMaximum(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 818 "QC_DatasourcePool.qpp"
    return ds->getMax();
}

// int DatasourcePool::getMinimum(){}
static QoreValue DatasourcePool_getMinimum(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 805 "QC_DatasourcePool.qpp"
    return ds->getMin();
}

// string DatasourcePool::getOSCharset(){}
static QoreValue DatasourcePool_getOSCharset(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 717 "QC_DatasourcePool.qpp"
    const QoreEncoding *enc = ds->getQoreEncoding();
   return new QoreStringNode(enc ? enc->getCode() : "(unknown)");
}

// *string DatasourcePool::getOSEncoding(){}
static QoreValue DatasourcePool_getOSEncoding(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 730 "QC_DatasourcePool.qpp"
    const QoreEncoding *enc = ds->getQoreEncoding();
   return enc ? new QoreStringNode(enc->getCode()) : 0;
}

// auto DatasourcePool::getOption(string opt){}
static QoreValue DatasourcePool_getOption_Vs(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* opt = HARD_QORE_VALUE_STRING(args, 0);
# 908 "QC_DatasourcePool.qpp"
    return ds->getOption(opt->getBuffer(), xsink);
}

// hash<auto> DatasourcePool::getOptionHash(){}
static QoreValue DatasourcePool_getOptionHash(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 897 "QC_DatasourcePool.qpp"
    return ds->getOptionHash();
}

// *string DatasourcePool::getPassword(){}
static QoreValue DatasourcePool_getPassword(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 663 "QC_DatasourcePool.qpp"
    return ds->getPendingPassword();
}

// *int DatasourcePool::getPort(){}
static QoreValue DatasourcePool_getPort(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 755 "QC_DatasourcePool.qpp"
    int port = ds->getPendingPort();
    return port ? QoreValue(port) : QoreValue();
}

// AbstractSQLStatement DatasourcePool::getSQLStatement(){}
static QoreValue DatasourcePool_getSQLStatement(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 1091 "QC_DatasourcePool.qpp"
    if (!ds->getDriver()->hasStatementAPI()) {
        xsink->raiseException("SQLSTATEMENT-ERROR", "DBI driver '%s' does not support the prepared statement API", ds->getDriver()->getName());
        return QoreValue();
     }

    return new QoreObject(QC_SQLSTATEMENT, getProgram(), new QoreSQLStatement(ds));
}

// auto DatasourcePool::getServerVersion(){}
static QoreValue DatasourcePool_getServerVersion(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 845 "QC_DatasourcePool.qpp"
    return ds->getServerVersion(xsink);
}

// *hash<auto> DatasourcePool::getUsageInfo(){}
static QoreValue DatasourcePool_getUsageInfo(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 991 "QC_DatasourcePool.qpp"
    return ds->getUsageInfo();
}

// *string DatasourcePool::getUserName(){}
static QoreValue DatasourcePool_getUserName(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 651 "QC_DatasourcePool.qpp"
    return ds->getPendingUsername();
}

// bool DatasourcePool::inTransaction(){}
static QoreValue DatasourcePool_inTransaction(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 871 "QC_DatasourcePool.qpp"
    return ds->inTransaction();
}

// nothing DatasourcePool::rollback(){}
static QoreValue DatasourcePool_rollback(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 383 "QC_DatasourcePool.qpp"
    ds->rollback(xsink);
    return QoreValue();
}

// auto DatasourcePool::select(string sql, ...){}
static QoreValue DatasourcePool_select_VsVV(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* sql = HARD_QORE_VALUE_STRING(args, 0);
# 481 "QC_DatasourcePool.qpp"
    ReferenceHolder<QoreListNode> vargs(args->size() > 1 ? args->copyListFrom(1) : 0, xsink);
   return ds->select(sql, *vargs, xsink);
}

// auto DatasourcePool::selectRow(string sql, ...){}
static QoreValue DatasourcePool_selectRow_VsVV(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* sql = HARD_QORE_VALUE_STRING(args, 0);
# 507 "QC_DatasourcePool.qpp"
    ReferenceHolder<QoreListNode> vargs(args->size() > 1 ? args->copyListFrom(1) : 0, xsink);
   return ds->selectRow(sql, *vargs, xsink);
}

// auto DatasourcePool::selectRows(string sql, ...){}
static QoreValue DatasourcePool_selectRows_VsVV(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* sql = HARD_QORE_VALUE_STRING(args, 0);
# 535 "QC_DatasourcePool.qpp"
    ReferenceHolder<QoreListNode> vargs(args->size() > 1 ? args->copyListFrom(1) : 0, xsink);
   return ds->selectRows(sql, *vargs, xsink);
}

// nothing DatasourcePool::setErrorTimeout(timeout ts){}
static QoreValue DatasourcePool_setErrorTimeout_Vt(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 ts = HARD_QORE_VALUE_INT(args, 0);
# 1003 "QC_DatasourcePool.qpp"
    if (ts < 0)
      ts = 0;
   ds->setErrorTimeout(ts);
    return QoreValue();
}

// nothing DatasourcePool::setEventQueue(Qore::Thread::Queue queue, auto arg){}
static QoreValue DatasourcePool_setEventQueue_C5QueueVA(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    HARD_QORE_VALUE_OBJ_DATA(queue, Queue, args, 0, CID_QUEUE, "DatasourcePool::setEventQueue()", "Qore::Thread::Queue", xsink);
    if (*xsink)
        return 0;
    QoreValue arg = get_param_value(args, 1);
# 1032 "QC_DatasourcePool.qpp"
    ReferenceHolder<Queue> q(queue, xsink);

   if (queue->getMax() != -1)
      return xsink->raiseException("QUEUE-ERROR", "the Queue object passed as a maximum size of %d entr%s, which could cause internal I/O to freeze in the Datasource object if used; use a Queue object with no maximum size for datasource event monitoring", queue->getMax(), queue->getMax() == 1 ? "y" : "ies");

   ds->setEventQueue(q.release(), arg.refSelf(), xsink);
    return QoreValue();
}

// nothing DatasourcePool::setWarningCallback(timeout ms, code callback, auto arg){}
static QoreValue DatasourcePool_setWarningCallback_VtVcVA(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 ms = HARD_QORE_VALUE_INT(args, 0);
    const ResolvedCallReferenceNode* callback = get_param_value(args, 1).get<const ResolvedCallReferenceNode>();
    QoreValue arg = get_param_value(args, 2);
# 969 "QC_DatasourcePool.qpp"
    ds->setWarningCallback(ms, callback->refRefSelf(), arg.refSelf(), xsink);
    return QoreValue();
}

// string DatasourcePool::toString(){}
static QoreValue DatasourcePool_toString(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 831 "QC_DatasourcePool.qpp"
    return ds->toString();
}

// auto DatasourcePool::vexec(string sql, *softlist<auto> vargs){}
static QoreValue DatasourcePool_vexec_VsC15_softlist_auto_(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* sql = HARD_QORE_VALUE_STRING(args, 0);
    const QoreListNode* vargs = get_param_value(args, 1).get<const QoreListNode>();
# 424 "QC_DatasourcePool.qpp"
    return ds->exec(sql, vargs, xsink);
}

// auto DatasourcePool::vselect(string sql, *softlist<auto> vargs){}
static QoreValue DatasourcePool_vselect_VsC15_softlist_auto_(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* sql = HARD_QORE_VALUE_STRING(args, 0);
    const QoreListNode* vargs = get_param_value(args, 1).get<const QoreListNode>();
# 569 "QC_DatasourcePool.qpp"
    return ds->select(sql, vargs, xsink);
}

// auto DatasourcePool::vselectRow(string sql, *softlist<auto> vargs){}
static QoreValue DatasourcePool_vselectRow_VsC15_softlist_auto_(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* sql = HARD_QORE_VALUE_STRING(args, 0);
    const QoreListNode* vargs = get_param_value(args, 1).get<const QoreListNode>();
# 594 "QC_DatasourcePool.qpp"
    return ds->selectRow(sql, vargs, xsink);
}

// auto DatasourcePool::vselectRows(string sql, *softlist<auto> vargs){}
static QoreValue DatasourcePool_vselectRows_VsC15_softlist_auto_(QoreObject* self, DatasourcePool* ds, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* sql = HARD_QORE_VALUE_STRING(args, 0);
    const QoreListNode* vargs = get_param_value(args, 1).get<const QoreListNode>();
# 623 "QC_DatasourcePool.qpp"
    return ds->selectRows(sql, vargs, xsink);
}

DLLLOCAL void preinitDatasourcePoolClass() {
    QC_DATASOURCEPOOL = new QoreClass("DatasourcePool", "::Qore::SQL::DatasourcePool", QDOM_DATABASE);
    CID_DATASOURCEPOOL = QC_DATASOURCEPOOL->getID();
    QC_DATASOURCEPOOL->setSystem();
}

DLLLOCAL QoreClass* initDatasourcePoolClass(QoreNamespace& ns) {
    if (!QC_DATASOURCEPOOL)
        preinitDatasourcePoolClass();

    // set serializer function
    QC_DATASOURCEPOOL->setSerializer(reinterpret_cast<q_serializer_t>(DatasourcePool_serializer));

    // set deserializer function
    QC_DATASOURCEPOOL->setDeserializer(DatasourcePool_deserializer);

    // set parent class
    assert(QC_ABSTRACTDATASOURCE);
    QC_DATASOURCEPOOL->addBuiltinVirtualBaseClass(QC_ABSTRACTDATASOURCE);

    // set parent class
    assert(QC_SERIALIZABLE);
    QC_DATASOURCEPOOL->addBuiltinVirtualBaseClass(QC_SERIALIZABLE);

    // nothing DatasourcePool::beginTransaction(){}
    QC_DATASOURCEPOOL->addMethod("beginTransaction", (q_method_n_t)DatasourcePool_beginTransaction, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

    // nothing DatasourcePool::clearEventQueue(){}
    QC_DATASOURCEPOOL->addMethod("clearEventQueue", (q_method_n_t)DatasourcePool_clearEventQueue, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

    // nothing DatasourcePool::clearWarningCallback(){}
    QC_DATASOURCEPOOL->addMethod("clearWarningCallback", (q_method_n_t)DatasourcePool_clearWarningCallback, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

    // nothing DatasourcePool::commit(){}
    QC_DATASOURCEPOOL->addMethod("commit", (q_method_n_t)DatasourcePool_commit, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

    // DatasourcePool::constructor(string driver, *string user, *string pass, *string db, *string encoding, *string host, softint min = 3, softint max = 10, softint port = 0, *Qore::Thread::Queue queue, auto arg) {}
    QC_DATASOURCEPOOL->addConstructor(DatasourcePool_constructor_VsNsNsNsNsNsviviviC5QueueVA, Public, QCF_NO_FLAGS, QDOM_DEFAULT, 11, stringTypeInfo, QORE_PARAM_NO_ARG, "driver", stringOrNothingTypeInfo, QORE_PARAM_NO_ARG, "user", stringOrNothingTypeInfo, QORE_PARAM_NO_ARG, "pass", stringOrNothingTypeInfo, QORE_PARAM_NO_ARG, "db", stringOrNothingTypeInfo, QORE_PARAM_NO_ARG, "encoding", stringOrNothingTypeInfo, QORE_PARAM_NO_ARG, "host", softBigIntTypeInfo, QoreSimpleValue().assign((int64)3), "min", softBigIntTypeInfo, QoreSimpleValue().assign((int64)10), "max", softBigIntTypeInfo, QoreSimpleValue().assign((int64)0), "port", QC_QUEUE->getOrNothingTypeInfo(), QORE_PARAM_NO_ARG, "queue", autoTypeInfo, QORE_PARAM_NO_ARG, "arg");

    // DatasourcePool::constructor(string desc, *Qore::Thread::Queue queue, auto arg) {}
    QC_DATASOURCEPOOL->addConstructor(DatasourcePool_constructor_VsC5QueueVA, Public, QCF_NO_FLAGS, QDOM_DEFAULT, 3, stringTypeInfo, QORE_PARAM_NO_ARG, "desc", QC_QUEUE->getOrNothingTypeInfo(), QORE_PARAM_NO_ARG, "queue", autoTypeInfo, QORE_PARAM_NO_ARG, "arg");

    // DatasourcePool::constructor(hash<auto> opts, *Qore::Thread::Queue queue, auto arg) {}
    QC_DATASOURCEPOOL->addConstructor(DatasourcePool_constructor_C10hash_auto_C5QueueVA, Public, QCF_NO_FLAGS, QDOM_DEFAULT, 3, autoHashTypeInfo, QORE_PARAM_NO_ARG, "opts", QC_QUEUE->getOrNothingTypeInfo(), QORE_PARAM_NO_ARG, "queue", autoTypeInfo, QORE_PARAM_NO_ARG, "arg");

    // DatasourcePool::copy() {}
    QC_DATASOURCEPOOL->setCopy((q_copy_t)DatasourcePool_copy);

    // bool DatasourcePool::currentThreadInTransaction(){}
    QC_DATASOURCEPOOL->addMethod("currentThreadInTransaction", (q_method_n_t)DatasourcePool_currentThreadInTransaction, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    // DatasourcePool::destructor() {}
    QC_DATASOURCEPOOL->setDestructor((q_destructor_t)DatasourcePool_destructor);

    // auto DatasourcePool::exec(string sql, ...){}
    QC_DATASOURCEPOOL->addMethod("exec", (q_method_n_t)DatasourcePool_exec_VsVV, Public, QCF_USES_EXTRA_ARGS, QDOM_DEFAULT, autoTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "sql");

    // auto DatasourcePool::execRaw(string sql){}
    QC_DATASOURCEPOOL->addMethod("execRaw", (q_method_n_t)DatasourcePool_execRaw_Vs, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "sql");

    // int DatasourcePool::getCapabilities(){}
    QC_DATASOURCEPOOL->addMethod("getCapabilities", (q_method_n_t)DatasourcePool_getCapabilities, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // list<auto> DatasourcePool::getCapabilityList(){}
    QC_DATASOURCEPOOL->addMethod("getCapabilityList", (q_method_n_t)DatasourcePool_getCapabilityList, Public, QCF_CONSTANT, QDOM_DEFAULT, autoListTypeInfo);

    // auto DatasourcePool::getClientVersion(){}
    QC_DATASOURCEPOOL->addMethod("getClientVersion", (q_method_n_t)DatasourcePool_getClientVersion, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo);

    // hash<auto> DatasourcePool::getConfigHash(){}
    QC_DATASOURCEPOOL->addMethod("getConfigHash", (q_method_n_t)DatasourcePool_getConfigHash, Public, QCF_CONSTANT, QDOM_DEFAULT, autoHashTypeInfo);

    // string DatasourcePool::getConfigString(){}
    QC_DATASOURCEPOOL->addMethod("getConfigString", (q_method_n_t)DatasourcePool_getConfigString, Public, QCF_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

    // *string DatasourcePool::getDBCharset(){}
    QC_DATASOURCEPOOL->addMethod("getDBCharset", (q_method_n_t)DatasourcePool_getDBCharset, Public, QCF_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // *string DatasourcePool::getDBEncoding(){}
    QC_DATASOURCEPOOL->addMethod("getDBEncoding", (q_method_n_t)DatasourcePool_getDBEncoding, Public, QCF_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // *string DatasourcePool::getDBName(){}
    QC_DATASOURCEPOOL->addMethod("getDBName", (q_method_n_t)DatasourcePool_getDBName, Public, QCF_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // string DatasourcePool::getDriverName(){}
    QC_DATASOURCEPOOL->addMethod("getDriverName", (q_method_n_t)DatasourcePool_getDriverName, Public, QCF_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

    // string DatasourcePool::getDriverRealName(){}
    QC_DATASOURCEPOOL->addMethod("getDriverRealName", (q_method_n_t)DatasourcePool_getDriverRealName, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo);

    // int DatasourcePool::getErrorTimeout(){}
    QC_DATASOURCEPOOL->addMethod("getErrorTimeout", (q_method_n_t)DatasourcePool_getErrorTimeout, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // *string DatasourcePool::getHostName(){}
    QC_DATASOURCEPOOL->addMethod("getHostName", (q_method_n_t)DatasourcePool_getHostName, Public, QCF_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // int DatasourcePool::getMaximum(){}
    QC_DATASOURCEPOOL->addMethod("getMaximum", (q_method_n_t)DatasourcePool_getMaximum, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // int DatasourcePool::getMinimum(){}
    QC_DATASOURCEPOOL->addMethod("getMinimum", (q_method_n_t)DatasourcePool_getMinimum, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // string DatasourcePool::getOSCharset(){}
    QC_DATASOURCEPOOL->addMethod("getOSCharset", (q_method_n_t)DatasourcePool_getOSCharset, Public, QCF_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

    // *string DatasourcePool::getOSEncoding(){}
    QC_DATASOURCEPOOL->addMethod("getOSEncoding", (q_method_n_t)DatasourcePool_getOSEncoding, Public, QCF_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // auto DatasourcePool::getOption(string opt){}
    QC_DATASOURCEPOOL->addMethod("getOption", (q_method_n_t)DatasourcePool_getOption_Vs, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, autoTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "opt");

    // hash<auto> DatasourcePool::getOptionHash(){}
    QC_DATASOURCEPOOL->addMethod("getOptionHash", (q_method_n_t)DatasourcePool_getOptionHash, Public, QCF_CONSTANT, QDOM_DEFAULT, autoHashTypeInfo);

    // *string DatasourcePool::getPassword(){}
    QC_DATASOURCEPOOL->addMethod("getPassword", (q_method_n_t)DatasourcePool_getPassword, Public, QCF_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // *int DatasourcePool::getPort(){}
    QC_DATASOURCEPOOL->addMethod("getPort", (q_method_n_t)DatasourcePool_getPort, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntOrNothingTypeInfo);

    // AbstractSQLStatement DatasourcePool::getSQLStatement(){}
    QC_DATASOURCEPOOL->addMethod("getSQLStatement", (q_method_n_t)DatasourcePool_getSQLStatement, Public, QCF_NO_FLAGS, QDOM_DEFAULT, QC_ABSTRACTSQLSTATEMENT->getTypeInfo());

    // auto DatasourcePool::getServerVersion(){}
    QC_DATASOURCEPOOL->addMethod("getServerVersion", (q_method_n_t)DatasourcePool_getServerVersion, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo);

    // *hash<auto> DatasourcePool::getUsageInfo(){}
    QC_DATASOURCEPOOL->addMethod("getUsageInfo", (q_method_n_t)DatasourcePool_getUsageInfo, Public, QCF_CONSTANT, QDOM_DEFAULT, autoHashOrNothingTypeInfo);

    // *string DatasourcePool::getUserName(){}
    QC_DATASOURCEPOOL->addMethod("getUserName", (q_method_n_t)DatasourcePool_getUserName, Public, QCF_CONSTANT, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // bool DatasourcePool::inTransaction(){}
    QC_DATASOURCEPOOL->addMethod("inTransaction", (q_method_n_t)DatasourcePool_inTransaction, Public, QCF_CONSTANT, QDOM_DEFAULT, boolTypeInfo);

    // nothing DatasourcePool::rollback(){}
    QC_DATASOURCEPOOL->addMethod("rollback", (q_method_n_t)DatasourcePool_rollback, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

    // auto DatasourcePool::select(string sql, ...){}
    QC_DATASOURCEPOOL->addMethod("select", (q_method_n_t)DatasourcePool_select_VsVV, Public, QCF_USES_EXTRA_ARGS, QDOM_DEFAULT, autoTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "sql");

    // auto DatasourcePool::selectRow(string sql, ...){}
    QC_DATASOURCEPOOL->addMethod("selectRow", (q_method_n_t)DatasourcePool_selectRow_VsVV, Public, QCF_USES_EXTRA_ARGS, QDOM_DEFAULT, autoTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "sql");

    // auto DatasourcePool::selectRows(string sql, ...){}
    QC_DATASOURCEPOOL->addMethod("selectRows", (q_method_n_t)DatasourcePool_selectRows_VsVV, Public, QCF_USES_EXTRA_ARGS, QDOM_DEFAULT, autoTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "sql");

    // nothing DatasourcePool::setErrorTimeout(timeout ts){}
    QC_DATASOURCEPOOL->addMethod("setErrorTimeout", (q_method_n_t)DatasourcePool_setErrorTimeout_Vt, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, timeoutTypeInfo, QORE_PARAM_NO_ARG, "ts");

    // nothing DatasourcePool::setEventQueue(Qore::Thread::Queue queue, auto arg){}
    QC_DATASOURCEPOOL->addMethod("setEventQueue", (q_method_n_t)DatasourcePool_setEventQueue_C5QueueVA, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, QC_QUEUE->getTypeInfo(), QORE_PARAM_NO_ARG, "queue", autoTypeInfo, QORE_PARAM_NO_ARG, "arg");

    // nothing DatasourcePool::setWarningCallback(timeout ms, code callback, auto arg){}
    QC_DATASOURCEPOOL->addMethod("setWarningCallback", (q_method_n_t)DatasourcePool_setWarningCallback_VtVcVA, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, timeoutTypeInfo, QORE_PARAM_NO_ARG, "ms", codeTypeInfo, QORE_PARAM_NO_ARG, "callback", autoTypeInfo, QORE_PARAM_NO_ARG, "arg");

    // string DatasourcePool::toString(){}
    QC_DATASOURCEPOOL->addMethod("toString", (q_method_n_t)DatasourcePool_toString, Public, QCF_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

    // auto DatasourcePool::vexec(string sql, *softlist<auto> vargs){}
    QC_DATASOURCEPOOL->addMethod("vexec", (q_method_n_t)DatasourcePool_vexec_VsC15_softlist_auto_, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "sql", softAutoListOrNothingTypeInfo, QORE_PARAM_NO_ARG, "vargs");

    // auto DatasourcePool::vselect(string sql, *softlist<auto> vargs){}
    QC_DATASOURCEPOOL->addMethod("vselect", (q_method_n_t)DatasourcePool_vselect_VsC15_softlist_auto_, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "sql", softAutoListOrNothingTypeInfo, QORE_PARAM_NO_ARG, "vargs");

    // auto DatasourcePool::vselectRow(string sql, *softlist<auto> vargs){}
    QC_DATASOURCEPOOL->addMethod("vselectRow", (q_method_n_t)DatasourcePool_vselectRow_VsC15_softlist_auto_, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "sql", softAutoListOrNothingTypeInfo, QORE_PARAM_NO_ARG, "vargs");

    // auto DatasourcePool::vselectRows(string sql, *softlist<auto> vargs){}
    QC_DATASOURCEPOOL->addMethod("vselectRows", (q_method_n_t)DatasourcePool_vselectRows_VsC15_softlist_auto_, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "sql", softAutoListOrNothingTypeInfo, QORE_PARAM_NO_ARG, "vargs");

    return QC_DATASOURCEPOOL;
}
