/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
    QC_ProgramControl.qpp

    Qore Programming Language

    Copyright (C) 2003 - 2023 Qore Technologies, s.r.o.

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

    Note that the Qore library is released under a choice of three open-source
    licenses: MIT (as above), LGPL 2+, or GPL 2+; see README-LICENSE for more
    information.
*/

#include <qore/Qore.h>
#include "qore/intern/QC_ProgramControl.h"
#include "qore/intern/QC_Breakpoint.h"
#include "qore/intern/QC_TimeZone.h"
#include "qore/intern/qore_program_private.h"
#include "qore/intern/QoreObjectIntern.h"
#include "qore/intern/qore_date_private.h"
#include "qore/intern/QoreHashNodeIntern.h"
TypedHashDecl* init_hashdecl_StatementInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("StatementInfo", "::Qore::StatementInfo");
    hd->addMember("breakpoints", listOrNothingTypeInfo, QoreValue());
    hd->addMember("end_line", bigIntTypeInfo, QoreValue());
    hd->addMember("file", stringTypeInfo, QoreValue());
    hd->addMember("offset", bigIntTypeInfo, QoreValue());
    hd->addMember("source", stringTypeInfo, QoreValue());
    hd->addMember("start_line", bigIntTypeInfo, QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

/* Qore class Qore::ProgramControl */

qore_classid_t CID_PROGRAMCONTROL;
QoreClass* QC_PROGRAMCONTROL;

// nothing ProgramControl::assignBreakpoint(Breakpoint bkpt){}
static QoreValue ProgramControl_assignBreakpoint_C10Breakpoint(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    HARD_QORE_VALUE_OBJ_DATA(bkpt, QoreBreakpoint, args, 0, CID_BREAKPOINT, "ProgramControl::assignBreakpoint()", "Breakpoint", xsink);
    if (*xsink)
        return 0;
# 267 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
      return QoreValue();
   p->assignBreakpoint(bkpt, xsink);
    return QoreValue();
}

// ProgramControl::constructor() {}
static void ProgramControl_constructor(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 83 "QC_ProgramControl.qpp"
xsink->raiseException("PROGRAM-COPY-ERROR", "creating ProgramControl objects is currently unsupported");
}

// ProgramControl::copy() {}
static void ProgramControl_copy(QoreObject* self, QoreObject* old, QoreProgram* p, ExceptionSink* xsink) {
# 91 "QC_ProgramControl.qpp"
xsink->raiseException("PROGRAM-COPY-ERROR", "copying ProgramControl objects is currently unsupported");
}

// nothing ProgramControl::deleteAllBreakpoints(){}
static QoreValue ProgramControl_deleteAllBreakpoints(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 276 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
      return QoreValue();
   p->deleteAllBreakpoints();
    return QoreValue();
}

// ProgramControl::destructor() {}
static void ProgramControl_destructor(QoreObject* self, QoreProgram* p, ExceptionSink* xsink) {
# 98 "QC_ProgramControl.qpp"
printd(5, "ProgramControl::constructor() pgm: %p, pgmid: %d, self: %p\n", p, p->getProgramId(), self);
   p->unregisterQoreObject(self, xsink);
   // the underlying object is created from C++ then it has reference and cannot be terminated
   p->deref(xsink);
}

// bool ProgramControl::existsFunction(string name){}
static QoreValue ProgramControl_existsFunction_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* name = HARD_QORE_VALUE_STRING(args, 0);
# 212 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
      return false;
   TempEncodingHelper tmp(name, QCS_DEFAULT, xsink);
   if (!tmp)
      return false;
   return p->existsFunction(tmp->getBuffer());
}

// int ProgramControl::findFunctionStatementId(string function, *list params){}
static QoreValue ProgramControl_findFunctionStatementId_VsNl(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* function = HARD_QORE_VALUE_STRING(args, 0);
    const QoreListNode* params = get_param_value(args, 1).get<const QoreListNode>();
# 362 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
        return 0;
    AbstractStatement* s = p->findFunctionStatement(function->c_str(), params, xsink);
    if (!s) {
        if (!*xsink) {
            xsink->raiseException("PROGRAM-STATEMENT-ERROR", "cannot find statement for \"%s()\"", function->c_str());
        }
        return QoreValue();
    }
    return p->getStatementId(s);
}

// list<hash<auto>> ProgramControl::findFunctionVariants(string function){}
static QoreValue ProgramControl_findFunctionVariants_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* function = HARD_QORE_VALUE_STRING(args, 0);
# 382 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
        return QoreValue();
    ReferenceHolder<QoreListNode> vl(p->runtimeFindCallVariants(function->c_str(), xsink), xsink);
    if (!vl)
        return xsink->raiseException("PROGRAM-ERROR", "cannot find any function or method \"%s\"", function->c_str());
    return vl.release();
}

// int ProgramControl::findStatementId(string file, int line){}
static QoreValue ProgramControl_findStatementId_VsVi(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* file = HARD_QORE_VALUE_STRING(args, 0);
    int64 line = HARD_QORE_VALUE_INT(args, 1);
# 344 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
        return 0;
    AbstractStatement* s = p->findStatement(file->c_str(), line);
    if (!s) {
        return xsink->raiseException("PROGRAM-STATEMENT-ERROR", "cannot find statement for \"%s:%d\"", file->c_str(), line);
    }
    return p->getStatementId(s);
}

// *hash<auto> ProgramControl::getAllDefines(){}
static QoreValue ProgramControl_getAllDefines(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 159 "QC_ProgramControl.qpp"
    return qore_program_private::runTimeGetAllDefines(p);
}

// list<Breakpoint> ProgramControl::getBreakpoints(){}
static QoreValue ProgramControl_getBreakpoints(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 285 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
        return QoreValue();
    QoreBreakpointList_t bl;
    p->getBreakpoints(bl);
    QoreListNode *l = new QoreListNode(QC_BREAKPOINT->getTypeInfo());
    for (QoreBreakpointList_t::iterator it = bl.begin(); it != bl.end(); ++it) {
        l->push((*it)->getQoreObject(), nullptr);
    }
    return l;
}

// auto ProgramControl::getDefine(string def){}
static QoreValue ProgramControl_getDefine_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* def = HARD_QORE_VALUE_STRING(args, 0);
# 151 "QC_ProgramControl.qpp"
    TempEncodingHelper t(def, QCS_DEFAULT, xsink);
   return t ? qore_program_private::runTimeGetDefine(p, t->c_str()) : QoreValue();
}

// auto ProgramControl::getGlobalVariable(string varname, *reference<bool> rexists){}
static QoreValue ProgramControl_getGlobalVariable_VsC16_reference_bool_(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* varname = HARD_QORE_VALUE_STRING(args, 0);
    const ReferenceNode* rexists = get_param_value(args, 1).get<const ReferenceNode>();
# 224 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
        return QoreValue();
    TempEncodingHelper t(varname, QCS_DEFAULT, xsink);
    if (!t)
        return QoreValue();

    bool found;
    ValueHolder rv(p->getGlobalVariableValue(t->c_str(), found), xsink);

    if (rexists) {
        QoreTypeSafeReferenceHelper r(rexists, xsink);
        if (!r)
            return QoreValue();

        if (r.assign(found))
            return QoreValue();
    }

    return rv.release();
}

// hash<auto> ProgramControl::getGlobalVars(){}
static QoreValue ProgramControl_getGlobalVars(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 249 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
      return QoreValue();
   return p->getGlobalVars();
}

// int ProgramControl::getParseOptions(){}
static QoreValue ProgramControl_getParseOptions(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 108 "QC_ProgramControl.qpp"
    return p->getParseOptions64();
}

// int ProgramControl::getProgramId(){}
static QoreValue ProgramControl_getProgramId(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 166 "QC_ProgramControl.qpp"
    return p->getProgramId();
}

// *string ProgramControl::getScriptDir(){}
static QoreValue ProgramControl_getScriptDir(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 115 "QC_ProgramControl.qpp"
    return p->getScriptDir();
}

// *string ProgramControl::getScriptName(){}
static QoreValue ProgramControl_getScriptName(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 122 "QC_ProgramControl.qpp"
    return p->getScriptName();
}

// *string ProgramControl::getScriptPath(){}
static QoreValue ProgramControl_getScriptPath(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 129 "QC_ProgramControl.qpp"
    return p->getScriptPath();
}

// hash<string,hash<string,int>> ProgramControl::getSourceFileNames(){}
static QoreValue ProgramControl_getSourceFileNames(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 417 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
      return QoreValue();
   return p->getSourceFileNames(xsink);
}

// hash<string,hash<string,int>> ProgramControl::getSourceLabels(){}
static QoreValue ProgramControl_getSourceLabels(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 401 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
      return QoreValue();
   return p->getSourceLabels(xsink);
}

// hash<StatementInfo> ProgramControl::getStatementIdInfo(int statementId, bool listBreakpoints = False){}
static QoreValue ProgramControl_getStatementIdInfo_ViVb(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 statementId = HARD_QORE_VALUE_INT(args, 0);
    bool listBreakpoints = HARD_QORE_VALUE_BOOL(args, 1);
# 308 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
      return QoreValue();
   AbstractStatement* s = p->resolveStatementId(statementId);
   if (!s) {
      return xsink->raiseException("PROGRAM-STATEMENT-ERROR", "cannot resolve statement id from (%d)", statementId);
   }
   QoreHashNode* h = new QoreHashNode(hashdeclStatementInfo, nullptr);
   qore_hash_private* ph = qore_hash_private::get(*h);
   ph->setKeyValueIntern("file", new QoreStringNode(s->loc->getFile()));
   ph->setKeyValueIntern("source", new QoreStringNode(s->loc->getSource()));
   ph->setKeyValueIntern("offset", s->loc->offset);
   ph->setKeyValueIntern("start_line", s->loc->start_line);
   ph->setKeyValueIntern("end_line", s->loc->end_line);
   if (listBreakpoints && p->checkAllowDebugging(xsink)) {
      QoreBreakpointList_t bl;
      p->getStatementBreakpoints(s, bl);
      QoreListNode *l = new QoreListNode(QC_BREAKPOINT->getTypeInfo());
      for (QoreBreakpointList_t::iterator it = bl.begin(); it != bl.end(); ++it) {
         l->push((*it)->getQoreObject(), xsink);
      }
      ph->setKeyValueIntern("breakpoints", l);
   }
   return h;
}

// list<int> ProgramControl::getThreadList(){}
static QoreValue ProgramControl_getThreadList(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 205 "QC_ProgramControl.qpp"
    return p->getThreadList();
}

// TimeZone ProgramControl::getTimeZone(){}
static QoreValue ProgramControl_getTimeZone(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 136 "QC_ProgramControl.qpp"
    return new QoreObject(QC_TIMEZONE, 0, new TimeZoneData(p->currentTZ()));
}

// bool ProgramControl::isDefined(string def){}
static QoreValue ProgramControl_isDefined_Vs(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* def = HARD_QORE_VALUE_STRING(args, 0);
# 143 "QC_ProgramControl.qpp"
    TempEncodingHelper t(def, QCS_DEFAULT, xsink);
   return t ? qore_program_private::runTimeIsDefined(p, t->getBuffer()) : false;
}

// nothing ProgramControl::setGlobalVarValue(string name, auto value){}
static QoreValue ProgramControl_setGlobalVarValue_VsVA(QoreObject* self, QoreProgram* p, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* name = HARD_QORE_VALUE_STRING(args, 0);
    QoreValue value = get_param_value(args, 1);
# 258 "QC_ProgramControl.qpp"
    if (!p->checkAllowDebugging(xsink))
      return QoreValue();
   p->setGlobalVarValue(name->c_str(), value.refSelf(), xsink);
    return QoreValue();
}

// static list<ProgramControl> ProgramControl::getAllPrograms(){}
static QoreValue static_ProgramControl_getAllPrograms(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 198 "QC_ProgramControl.qpp"
return QoreProgram::getAllQoreObjects(xsink);
}

// static ProgramControl ProgramControl::getProgram(){}
static QoreValue static_ProgramControl_getProgram(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 189 "QC_ProgramControl.qpp"
QoreProgram* pgm = getProgram();
   return QoreProgram::getQoreObject(pgm);
}

// static ProgramControl ProgramControl::resolveProgramId(int programId){}
static QoreValue static_ProgramControl_resolveProgramId_Vi(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 programId = HARD_QORE_VALUE_INT(args, 0);
# 178 "QC_ProgramControl.qpp"
QoreProgram* pgm = QoreProgram::resolveProgramId(programId);
   if (!pgm) {
      return xsink->raiseException("PROGRAM-ERROR", "cannot find program for programId (%d)", programId);
   }
   return QoreProgram::getQoreObject(pgm);
}

DLLLOCAL void preinitProgramControlClass() {
    QC_PROGRAMCONTROL = new QoreClass("ProgramControl", "::Qore::ProgramControl", QDOM_DEBUGGER);
    CID_PROGRAMCONTROL = QC_PROGRAMCONTROL->getID();
    QC_PROGRAMCONTROL->setSystem();
}

DLLLOCAL QoreClass* initProgramControlClass(QoreNamespace& ns) {
    if (!QC_PROGRAMCONTROL)
        preinitProgramControlClass();

    // nothing ProgramControl::assignBreakpoint(Breakpoint bkpt){}
    QC_PROGRAMCONTROL->addMethod("assignBreakpoint", (q_method_n_t)ProgramControl_assignBreakpoint_C10Breakpoint, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, QC_BREAKPOINT->getTypeInfo(), QORE_PARAM_NO_ARG, "bkpt");

    // ProgramControl::constructor() {}
    QC_PROGRAMCONTROL->addConstructor(ProgramControl_constructor, Public, QCF_NO_FLAGS, QDOM_DEFAULT);

    // ProgramControl::copy() {}
    QC_PROGRAMCONTROL->setCopy((q_copy_t)ProgramControl_copy);

    // nothing ProgramControl::deleteAllBreakpoints(){}
    QC_PROGRAMCONTROL->addMethod("deleteAllBreakpoints", (q_method_n_t)ProgramControl_deleteAllBreakpoints, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

    // ProgramControl::destructor() {}
    QC_PROGRAMCONTROL->setDestructor((q_destructor_t)ProgramControl_destructor);

    // bool ProgramControl::existsFunction(string name){}
    QC_PROGRAMCONTROL->addMethod("existsFunction", (q_method_n_t)ProgramControl_existsFunction_Vs, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, boolTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "name");

    // int ProgramControl::findFunctionStatementId(string function, *list params){}
    QC_PROGRAMCONTROL->addMethod("findFunctionStatementId", (q_method_n_t)ProgramControl_findFunctionStatementId_VsNl, Public, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "function", listOrNothingTypeInfo, QORE_PARAM_NO_ARG, "params");

    // list<hash<auto>> ProgramControl::findFunctionVariants(string function){}
    QC_PROGRAMCONTROL->addMethod("findFunctionVariants", (q_method_n_t)ProgramControl_findFunctionVariants_Vs, Public, QCF_NO_FLAGS, QDOM_DEFAULT, qore_get_complex_list_type(autoHashTypeInfo), 1, stringTypeInfo, QORE_PARAM_NO_ARG, "function");

    // int ProgramControl::findStatementId(string file, int line){}
    QC_PROGRAMCONTROL->addMethod("findStatementId", (q_method_n_t)ProgramControl_findStatementId_VsVi, Public, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "file", bigIntTypeInfo, QORE_PARAM_NO_ARG, "line");

    // *hash<auto> ProgramControl::getAllDefines(){}
    QC_PROGRAMCONTROL->addMethod("getAllDefines", (q_method_n_t)ProgramControl_getAllDefines, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoHashOrNothingTypeInfo);

    // list<Breakpoint> ProgramControl::getBreakpoints(){}
    QC_PROGRAMCONTROL->addMethod("getBreakpoints", (q_method_n_t)ProgramControl_getBreakpoints, Public, QCF_NO_FLAGS, QDOM_DEFAULT, qore_get_complex_list_type(QC_BREAKPOINT->getTypeInfo()));

    // auto ProgramControl::getDefine(string def){}
    QC_PROGRAMCONTROL->addMethod("getDefine", (q_method_n_t)ProgramControl_getDefine_Vs, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "def");

    // auto ProgramControl::getGlobalVariable(string varname, *reference<bool> rexists){}
    QC_PROGRAMCONTROL->addMethod("getGlobalVariable", (q_method_n_t)ProgramControl_getGlobalVariable_VsC16_reference_bool_, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "varname", qore_get_complex_reference_or_nothing_type(boolTypeInfo), QORE_PARAM_NO_ARG, "rexists");

    // hash<auto> ProgramControl::getGlobalVars(){}
    QC_PROGRAMCONTROL->addMethod("getGlobalVars", (q_method_n_t)ProgramControl_getGlobalVars, Public, QCF_CONSTANT, QDOM_DEFAULT, autoHashTypeInfo);

    // int ProgramControl::getParseOptions(){}
    QC_PROGRAMCONTROL->addMethod("getParseOptions", (q_method_n_t)ProgramControl_getParseOptions, Public, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);

    // int ProgramControl::getProgramId(){}
    QC_PROGRAMCONTROL->addMethod("getProgramId", (q_method_n_t)ProgramControl_getProgramId, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, bigIntTypeInfo);

    // *string ProgramControl::getScriptDir(){}
    QC_PROGRAMCONTROL->addMethod("getScriptDir", (q_method_n_t)ProgramControl_getScriptDir, Public, QCF_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // *string ProgramControl::getScriptName(){}
    QC_PROGRAMCONTROL->addMethod("getScriptName", (q_method_n_t)ProgramControl_getScriptName, Public, QCF_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // *string ProgramControl::getScriptPath(){}
    QC_PROGRAMCONTROL->addMethod("getScriptPath", (q_method_n_t)ProgramControl_getScriptPath, Public, QCF_NO_FLAGS, QDOM_DEFAULT, stringOrNothingTypeInfo);

    // hash<string,hash<string,int>> ProgramControl::getSourceFileNames(){}
    QC_PROGRAMCONTROL->addMethod("getSourceFileNames", (q_method_n_t)ProgramControl_getSourceFileNames, Public, QCF_NO_FLAGS, QDOM_DEFAULT, qore_get_complex_hash_type(qore_get_complex_hash_type(bigIntTypeInfo)));

    // hash<string,hash<string,int>> ProgramControl::getSourceLabels(){}
    QC_PROGRAMCONTROL->addMethod("getSourceLabels", (q_method_n_t)ProgramControl_getSourceLabels, Public, QCF_NO_FLAGS, QDOM_DEFAULT, qore_get_complex_hash_type(qore_get_complex_hash_type(bigIntTypeInfo)));

    // hash<StatementInfo> ProgramControl::getStatementIdInfo(int statementId, bool listBreakpoints = False){}
    QC_PROGRAMCONTROL->addMethod("getStatementIdInfo", (q_method_n_t)ProgramControl_getStatementIdInfo_ViVb, Public, QCF_NO_FLAGS, QDOM_DEFAULT, hashdeclStatementInfo->getTypeInfo(false), 2, bigIntTypeInfo, QORE_PARAM_NO_ARG, "statementId", boolTypeInfo, QoreSimpleValue().assign(false), "listBreakpoints");

    // list<int> ProgramControl::getThreadList(){}
    QC_PROGRAMCONTROL->addMethod("getThreadList", (q_method_n_t)ProgramControl_getThreadList, Public, QCF_CONSTANT, QDOM_DEFAULT, qore_get_complex_list_type(bigIntTypeInfo));

    // TimeZone ProgramControl::getTimeZone(){}
    QC_PROGRAMCONTROL->addMethod("getTimeZone", (q_method_n_t)ProgramControl_getTimeZone, Public, QCF_NO_FLAGS, QDOM_DEFAULT, QC_TIMEZONE->getTypeInfo());

    // bool ProgramControl::isDefined(string def){}
    QC_PROGRAMCONTROL->addMethod("isDefined", (q_method_n_t)ProgramControl_isDefined_Vs, Public, QCF_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "def");

    // nothing ProgramControl::setGlobalVarValue(string name, auto value){}
    QC_PROGRAMCONTROL->addMethod("setGlobalVarValue", (q_method_n_t)ProgramControl_setGlobalVarValue_VsVA, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "name", autoTypeInfo, QORE_PARAM_NO_ARG, "value");

    // static list<ProgramControl> ProgramControl::getAllPrograms(){}
    QC_PROGRAMCONTROL->addStaticMethod("getAllPrograms", (q_func_n_t)static_ProgramControl_getAllPrograms, Public, QCF_RET_VALUE_ONLY, QDOM_DEBUGGER, qore_get_complex_list_type(QC_PROGRAMCONTROL->getTypeInfo()));

    // static ProgramControl ProgramControl::getProgram(){}
    QC_PROGRAMCONTROL->addStaticMethod("getProgram", (q_func_n_t)static_ProgramControl_getProgram, Public, QCF_RET_VALUE_ONLY, QDOM_DEBUGGER, QC_PROGRAMCONTROL->getTypeInfo());

    // static ProgramControl ProgramControl::resolveProgramId(int programId){}
    QC_PROGRAMCONTROL->addStaticMethod("resolveProgramId", (q_func_n_t)static_ProgramControl_resolveProgramId_Vi, Public, QCF_NO_FLAGS, QDOM_DEBUGGER, QC_PROGRAMCONTROL->getTypeInfo(), 1, bigIntTypeInfo, QORE_PARAM_NO_ARG, "programId");

    return QC_PROGRAMCONTROL;
}
