/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
    QC_SSLPrivateKey.qpp

    Qore Programming Language

    Copyright (C) 2003 - 2023 Qore Technologies, s.r.o.

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

    Note that the Qore library is released under a choice of three open-source
    licenses: MIT (as above), LGPL 2+, or GPL 2+; see README-LICENSE for more
    information.
*/

#include <qore/Qore.h>
#include "qore/intern/QC_SSLPrivateKey.h"
#include "qore/intern/QC_Serializable.h"

#include <openssl/ssl.h>
#include <openssl/pem.h>

static QoreHashNode* SSLPrivateKey_serializer(const QoreObject& self, const QoreSSLPrivateKey& pk, QoreSerializationContext& context, ExceptionSink* xsink) {
    ReferenceHolder<QoreHashNode> rv(new QoreHashNode(autoTypeInfo), xsink);
    // we serialize with DER format as it's approx 2/3 the size of a PEM string
    SimpleRefHolder<BinaryNode> der(pk.getDER(xsink));
    if (*xsink) {
        return nullptr;
    }
    //printd(5, "SSLPrivateKey_serializer() der: %p %d bytes\n", *der, der->size());
    rv->setKeyValue("der", der.release(), xsink);
    return *xsink ? nullptr : rv.release();
}

static void SSLPrivateKey_deserializer(QoreObject& self, const QoreHashNode* sdata, QoreDeserializationContext& context, ExceptionSink* xsink) {
    if (sdata) {
        // we serialize with DER format as it's approx 2/3 the size of a PEM string
        QoreValue val = sdata->getKeyValue("der");
        if (val) {
            if (val.getType() != NT_BINARY) {
                xsink->raiseException("DESERIALIZATION-ERROR", "SSLPrivateKey serialization key 'der' has type '%s'; expecting 'binary'", val.getTypeName());
                return;
            }
            const BinaryNode* der = val.get<const BinaryNode>();
            SimpleRefHolder<QoreSSLPrivateKey> qpk(new QoreSSLPrivateKey(der, xsink));
            if (!*xsink) {
                self.setPrivate(CID_SSLPRIVATEKEY, qpk.release());
            }
            return;
        }
    }

    xsink->raiseException("DESERIALIZATION-ERROR", "missing serialization data for the SSLPrivateKey class");
}
/* Qore class Qore::SSLPrivateKey */

qore_classid_t CID_SSLPRIVATEKEY;
QoreClass* QC_SSLPRIVATEKEY;

// SSLPrivateKey::constructor(string pem, *string pass) {}
static void SSLPrivateKey_constructor_VsNs(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* pem = HARD_QORE_VALUE_STRING(args, 0);
    const QoreStringNode* pass = get_param_value(args, 1).get<const QoreStringNode>();
# 94 "QC_SSLPrivateKey.qpp"
SimpleRefHolder<QoreSSLPrivateKey> qpk(new QoreSSLPrivateKey(pem, pass ? pass->c_str() : nullptr, xsink));

    if (!*xsink) {
        self->setPrivate(CID_SSLPRIVATEKEY, qpk.release());
    }
}

// SSLPrivateKey::constructor(binary der) {}
static void SSLPrivateKey_constructor_Vy(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const BinaryNode* der = HARD_QORE_VALUE_BINARY(args, 0);
# 112 "QC_SSLPrivateKey.qpp"
SimpleRefHolder<QoreSSLPrivateKey> qc(new QoreSSLPrivateKey(der, xsink));
    if (!*xsink) {
        self->setPrivate(CID_SSLPRIVATEKEY, qc.release());
    }
}

// SSLPrivateKey::copy() {}
static void SSLPrivateKey_copy(QoreObject* self, QoreObject* old, QoreSSLPrivateKey* pk, ExceptionSink* xsink) {
# 123 "QC_SSLPrivateKey.qpp"
SimpleRefHolder<BinaryNode> der(pk->getDER(xsink));
    if (*xsink) {
        return;
    }
    SimpleRefHolder<QoreSSLPrivateKey> qpk(new QoreSSLPrivateKey(*der, xsink));
    if (!*xsink) {
        self->setPrivate(CID_SSLPRIVATEKEY, qpk.release());
    }
}

// int SSLPrivateKey::getBitLength(){}
static QoreValue SSLPrivateKey_getBitLength(QoreObject* self, QoreSSLPrivateKey* pk, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 178 "QC_SSLPrivateKey.qpp"
    return pk->getBitLength();
}

// binary SSLPrivateKey::getDER(){}
static QoreValue SSLPrivateKey_getDER(QoreObject* self, QoreSSLPrivateKey* pk, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 214 "QC_SSLPrivateKey.qpp"
    return pk->getDER(xsink);
}

// hash<auto> SSLPrivateKey::getInfo(){}
static QoreValue SSLPrivateKey_getInfo(QoreObject* self, QoreSSLPrivateKey* pk, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 145 "QC_SSLPrivateKey.qpp"
    return pk->getInfo();
}

// string SSLPrivateKey::getPEM(){}
static QoreValue SSLPrivateKey_getPEM(QoreObject* self, QoreSSLPrivateKey* pk, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 195 "QC_SSLPrivateKey.qpp"
    return pk->getPEM(xsink);
}

// string SSLPrivateKey::getType(){}
static QoreValue SSLPrivateKey_getType(QoreObject* self, QoreSSLPrivateKey* pk, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 157 "QC_SSLPrivateKey.qpp"
    return new QoreStringNode(pk->getType());
}

// int SSLPrivateKey::getVersion(){}
static QoreValue SSLPrivateKey_getVersion(QoreObject* self, QoreSSLPrivateKey* pk, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 166 "QC_SSLPrivateKey.qpp"
    return pk->getVersion();
}

DLLLOCAL void preinitSSLPrivateKeyClass() {
    QC_SSLPRIVATEKEY = new QoreClass("SSLPrivateKey", "::Qore::SSLPrivateKey", QDOM_DEFAULT);
    CID_SSLPRIVATEKEY = QC_SSLPRIVATEKEY->getID();
    QC_SSLPRIVATEKEY->setSystem();
}

DLLLOCAL QoreClass* initSSLPrivateKeyClass(QoreNamespace& ns) {
    if (!QC_SSLPRIVATEKEY)
        preinitSSLPrivateKeyClass();

    // set serializer function
    QC_SSLPRIVATEKEY->setSerializer(reinterpret_cast<q_serializer_t>(SSLPrivateKey_serializer));

    // set deserializer function
    QC_SSLPRIVATEKEY->setDeserializer(SSLPrivateKey_deserializer);

    // set parent class
    assert(QC_SERIALIZABLE);
    QC_SSLPRIVATEKEY->addBuiltinVirtualBaseClass(QC_SERIALIZABLE);

    // SSLPrivateKey::constructor(string pem, *string pass) {}
    QC_SSLPRIVATEKEY->addConstructor(SSLPrivateKey_constructor_VsNs, Public, QCF_NO_FLAGS, QDOM_DEFAULT, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "pem", stringOrNothingTypeInfo, QORE_PARAM_NO_ARG, "pass");

    // SSLPrivateKey::constructor(binary der) {}
    QC_SSLPRIVATEKEY->addConstructor(SSLPrivateKey_constructor_Vy, Public, QCF_NO_FLAGS, QDOM_DEFAULT, 1, binaryTypeInfo, QORE_PARAM_NO_ARG, "der");

    // SSLPrivateKey::copy() {}
    QC_SSLPRIVATEKEY->setCopy((q_copy_t)SSLPrivateKey_copy);

    // int SSLPrivateKey::getBitLength(){}
    QC_SSLPRIVATEKEY->addMethod("getBitLength", (q_method_n_t)SSLPrivateKey_getBitLength, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    // binary SSLPrivateKey::getDER(){}
    QC_SSLPRIVATEKEY->addMethod("getDER", (q_method_n_t)SSLPrivateKey_getDER, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, binaryTypeInfo);

    // hash<auto> SSLPrivateKey::getInfo(){}
    QC_SSLPRIVATEKEY->addMethod("getInfo", (q_method_n_t)SSLPrivateKey_getInfo, Public, QCF_CONSTANT, QDOM_DEFAULT, autoHashTypeInfo);

    // string SSLPrivateKey::getPEM(){}
    QC_SSLPRIVATEKEY->addMethod("getPEM", (q_method_n_t)SSLPrivateKey_getPEM, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo);

    // string SSLPrivateKey::getType(){}
    QC_SSLPRIVATEKEY->addMethod("getType", (q_method_n_t)SSLPrivateKey_getType, Public, QCF_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

    // int SSLPrivateKey::getVersion(){}
    QC_SSLPRIVATEKEY->addMethod("getVersion", (q_method_n_t)SSLPrivateKey_getVersion, Public, QCF_CONSTANT, QDOM_DEFAULT, bigIntTypeInfo);

    return QC_SSLPRIVATEKEY;
}
