/* -*- mode: c++; indent-tabs-mode: nil -*- */
/*
    QC_Serializable.qpp

    Qore Programming Language

    Copyright (C) 2003 - 2023 Qore Technologies, s.r.o.

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

    Note that the Qore library is released under a choice of three open-source
    licenses: MIT (as above), LGPL 2+, or GPL 2+; see README-LICENSE for more
    information.
*/

#include <qore/Qore.h>
#include "qore/intern/QC_Serializable.h"
#include "qore/intern/BinaryOutputStream.h"
#include "qore/intern/BinaryInputStream.h"
#include "qore/intern/StringInputStream.h"
#include "qore/intern/QC_TimeZone.h"

static QoreHashNode* Serializable_serializer(const QoreObject& self, const TimeZoneData& z, QoreSerializationContext& context, ExceptionSink* xsink) {
    return nullptr;
}

static void Serializable_deserializer(QoreObject& self, const QoreHashNode* sdata, QoreDeserializationContext& context, ExceptionSink* xsink) {
    if (sdata) {
        xsink->raiseException("DESERIALIZATION-ERROR", "deserialization data provided for the 'Serializable' class where none was expected");
        return;
    }
    self.setPrivate(CID_SERIALIZABLE, new QoreSerializable);
}
TypedHashDecl* init_hashdecl_SerializationInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("SerializationInfo", "::Qore::SerializationInfo");
    hd->addMember("_data", autoTypeInfo, QoreValue());
    hd->addMember("_index", qore_get_complex_hash_or_nothing_type(hashdeclObjectSerializationInfo->getTypeInfo(false)), QoreValue());
    hd->addMember("_modules", qore_get_complex_list_or_nothing_type(stringTypeInfo), QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

TypedHashDecl* init_hashdecl_ObjectSerializationInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("ObjectSerializationInfo", "::Qore::ObjectSerializationInfo");
    hd->addMember("_class", stringTypeInfo, QoreValue());
    hd->addMember("_class_data", qore_get_complex_hash_or_nothing_type(qore_get_complex_hash_type(autoTypeInfo)), QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

TypedHashDecl* init_hashdecl_IndexedObjectSerializationInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("IndexedObjectSerializationInfo", "::Qore::IndexedObjectSerializationInfo");
    hd->addMember("_index", stringTypeInfo, QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

TypedHashDecl* init_hashdecl_HashSerializationInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("HashSerializationInfo", "::Qore::HashSerializationInfo");
    hd->addMember("_hash", stringTypeInfo, QoreValue());
    hd->addMember("_members", qore_get_complex_hash_or_nothing_type(autoTypeInfo), QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

TypedHashDecl* init_hashdecl_ListSerializationInfo(QoreNamespace& ns) {
    TypedHashDecl* hd = new TypedHashDecl("ListSerializationInfo", "::Qore::ListSerializationInfo");
    hd->addMember("_elements", autoListOrNothingTypeInfo, QoreValue());
    hd->addMember("_list", stringTypeInfo, QoreValue());
    ns.addSystemHashDecl(hd);
    return hd;
}

/* Qore class Qore::Serializable */

qore_classid_t CID_SERIALIZABLE;
QoreClass* QC_SERIALIZABLE;

// Serializable::constructor() {}
static void Serializable_constructor(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 139 "QC_Serializable.qpp"
self->setPrivate(CID_SERIALIZABLE, new QoreSerializable);
}

// Serializable::copy() {}
static void Serializable_copy(QoreObject* self, QoreObject* old, QoreSerializable* s, ExceptionSink* xsink) {
# 146 "QC_Serializable.qpp"
self->setPrivate(CID_SERIALIZABLE, new QoreSerializable);
}

// private:internal nothing Serializable::deserializeMembers(hash<auto> members){}
static QoreValue Serializable_deserializeMembers_C10hash_auto_(QoreObject* self, QoreSerializable* s, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 298 "QC_Serializable.qpp"
    
    return QoreValue();
}

// nothing Serializable::serialize(OutputStream stream){}
static QoreValue Serializable_serialize_C12OutputStream(QoreObject* self, QoreSerializable* s, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    HARD_QORE_VALUE_OBJ_DATA(stream, OutputStream, args, 0, CID_OUTPUTSTREAM, "Serializable::serialize()", "OutputStream", xsink);
    if (*xsink)
        return 0;
# 186 "QC_Serializable.qpp"
    ReferenceHolder<OutputStream> holder(stream, xsink);
    QoreSerializable::serialize(*self, *stream, xsink);
    return QoreValue();
}

// binary Serializable::serialize(){}
static QoreValue Serializable_serialize(QoreObject* self, QoreSerializable* s, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 207 "QC_Serializable.qpp"
    ReferenceHolder<BinaryOutputStream> stream(new BinaryOutputStream, xsink);
    QoreSerializable::serialize(*self, **stream, xsink);
    return *xsink ? QoreValue() : stream->getData(xsink);
}

// private:internal *hash<auto> Serializable::serializeMembers(*hash<auto> members){}
static QoreValue Serializable_serializeMembers_C11_hash_auto_(QoreObject* self, QoreSerializable* s, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 256 "QC_Serializable.qpp"
    
    return QoreValue();
}

// hash<SerializationInfo> Serializable::serializeToData(){}
static QoreValue Serializable_serializeToData(QoreObject* self, QoreSerializable* s, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 166 "QC_Serializable.qpp"
    return s->serializeToData(self, xsink);
}

// static auto Serializable::deserialize(InputStream stream){}
static QoreValue static_Serializable_deserialize_C11InputStream(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    HARD_QORE_VALUE_OBJ_DATA(stream, InputStream, args, 0, CID_INPUTSTREAM, "Serializable::deserialize()", "InputStream", xsink);
    if (*xsink)
        return 0;
# 347 "QC_Serializable.qpp"
ReferenceHolder<InputStream> holder(stream, xsink);
    return QoreSerializable::deserialize(*stream, xsink);
}

// static auto Serializable::deserialize(binary bin){}
static QoreValue static_Serializable_deserialize_Vy(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const BinaryNode* bin = HARD_QORE_VALUE_BINARY(args, 0);
# 363 "QC_Serializable.qpp"
ReferenceHolder<BinaryInputStream> stream(new BinaryInputStream(bin), xsink);
    return QoreSerializable::deserialize(**stream, xsink);
}

// static auto Serializable::deserialize(string bin){}
static QoreValue static_Serializable_deserialize_Vs(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* bin = HARD_QORE_VALUE_STRING(args, 0);
# 382 "QC_Serializable.qpp"
ReferenceHolder<StringInputStream> stream(new StringInputStream(bin), xsink);
    return QoreSerializable::deserialize(**stream, xsink);
}

// static auto Serializable::deserialize(hash<SerializationInfo> data){}
static QoreValue static_Serializable_deserialize_C23hash_SerializationInfo_(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreHashNode* data = HARD_QORE_VALUE_HASH(args, 0);
# 399 "QC_Serializable.qpp"
return QoreSerializable::deserialize(*data, xsink);
}

// static hash<SerializationInfo> Serializable::deserializeToData(InputStream stream){}
static QoreValue static_Serializable_deserializeToData_C11InputStream(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    HARD_QORE_VALUE_OBJ_DATA(stream, InputStream, args, 0, CID_INPUTSTREAM, "Serializable::deserializeToData()", "InputStream", xsink);
    if (*xsink)
        return 0;
# 313 "QC_Serializable.qpp"
ReferenceHolder<InputStream> holder(stream, xsink);
    return QoreSerializable::deserializeToData(*stream, xsink);
}

// static hash<SerializationInfo> Serializable::deserializeToData(binary bin){}
static QoreValue static_Serializable_deserializeToData_Vy(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const BinaryNode* bin = HARD_QORE_VALUE_BINARY(args, 0);
# 330 "QC_Serializable.qpp"
ReferenceHolder<BinaryInputStream> stream(new BinaryInputStream(bin), xsink);
    return QoreSerializable::deserializeToData(**stream, xsink);
}

// static nothing Serializable::serialize(auto val, OutputStream stream){}
static QoreValue static_Serializable_serialize_VAC12OutputStream(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    QoreValue val = get_param_value(args, 0);
    HARD_QORE_VALUE_OBJ_DATA(stream, OutputStream, args, 1, CID_OUTPUTSTREAM, "Serializable::serialize()", "OutputStream", xsink);
    if (*xsink)
        return 0;
# 442 "QC_Serializable.qpp"
ReferenceHolder<OutputStream> holder(stream, xsink);
    QoreSerializable::serialize(val, *stream, xsink);
    return QoreValue();
}

// static binary Serializable::serialize(auto val){}
static QoreValue static_Serializable_serialize_VA(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    QoreValue val = get_param_value(args, 0);
# 465 "QC_Serializable.qpp"
ReferenceHolder<BinaryOutputStream> stream(new BinaryOutputStream, xsink);
    QoreSerializable::serialize(val, **stream, xsink);
    return *xsink ? QoreValue() : stream->getData(xsink);
}

// static hash<SerializationInfo> Serializable::serializeToData(auto val){}
static QoreValue static_Serializable_serializeToData_VA(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    QoreValue val = get_param_value(args, 0);
# 421 "QC_Serializable.qpp"
return QoreSerializable::serializeToData(val, xsink);
}

DLLLOCAL void preinitSerializableClass() {
    QC_SERIALIZABLE = new QoreClass("Serializable", "::Qore::Serializable", QDOM_DEFAULT);
    CID_SERIALIZABLE = QC_SERIALIZABLE->getID();
    QC_SERIALIZABLE->setSystem();
}

DLLLOCAL QoreClass* initSerializableClass(QoreNamespace& ns) {
    if (!QC_SERIALIZABLE)
        preinitSerializableClass();

    // set serializer function
    QC_SERIALIZABLE->setSerializer(reinterpret_cast<q_serializer_t>(Serializable_serializer));

    // set deserializer function
    QC_SERIALIZABLE->setDeserializer(Serializable_deserializer);

    // Serializable::constructor() {}
    QC_SERIALIZABLE->addConstructor(Serializable_constructor, Public, QCF_NO_FLAGS, QDOM_DEFAULT);

    // Serializable::copy() {}
    QC_SERIALIZABLE->setCopy((q_copy_t)Serializable_copy);

    // private:internal nothing Serializable::deserializeMembers(hash<auto> members){}
    QC_SERIALIZABLE->addMethod("deserializeMembers", (q_method_n_t)Serializable_deserializeMembers_C10hash_auto_, Internal, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, autoHashTypeInfo, QORE_PARAM_NO_ARG, "members");

    // nothing Serializable::serialize(OutputStream stream){}
    QC_SERIALIZABLE->addMethod("serialize", (q_method_n_t)Serializable_serialize_C12OutputStream, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, QC_OUTPUTSTREAM->getTypeInfo(), QORE_PARAM_NO_ARG, "stream");

    // binary Serializable::serialize(){}
    QC_SERIALIZABLE->addMethod("serialize", (q_method_n_t)Serializable_serialize, Public, QCF_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo);

    // private:internal *hash<auto> Serializable::serializeMembers(*hash<auto> members){}
    QC_SERIALIZABLE->addMethod("serializeMembers", (q_method_n_t)Serializable_serializeMembers_C11_hash_auto_, Internal, QCF_NO_FLAGS, QDOM_DEFAULT, autoHashOrNothingTypeInfo, 1, autoHashOrNothingTypeInfo, QORE_PARAM_NO_ARG, "members");

    // hash<SerializationInfo> Serializable::serializeToData(){}
    QC_SERIALIZABLE->addMethod("serializeToData", (q_method_n_t)Serializable_serializeToData, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, hashdeclSerializationInfo->getTypeInfo(false));

    // static auto Serializable::deserialize(InputStream stream){}
    QC_SERIALIZABLE->addStaticMethod("deserialize", (q_func_n_t)static_Serializable_deserialize_C11InputStream, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, autoTypeInfo, 1, QC_INPUTSTREAM->getTypeInfo(), QORE_PARAM_NO_ARG, "stream");

    // static auto Serializable::deserialize(binary bin){}
    QC_SERIALIZABLE->addStaticMethod("deserialize", (q_func_n_t)static_Serializable_deserialize_Vy, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, autoTypeInfo, 1, binaryTypeInfo, QORE_PARAM_NO_ARG, "bin");

    // static auto Serializable::deserialize(string bin){}
    QC_SERIALIZABLE->addStaticMethod("deserialize", (q_func_n_t)static_Serializable_deserialize_Vs, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, autoTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "bin");

    // static auto Serializable::deserialize(hash<SerializationInfo> data){}
    QC_SERIALIZABLE->addStaticMethod("deserialize", (q_func_n_t)static_Serializable_deserialize_C23hash_SerializationInfo_, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, autoTypeInfo, 1, hashdeclSerializationInfo->getTypeInfo(false), QORE_PARAM_NO_ARG, "data");

    // static hash<SerializationInfo> Serializable::deserializeToData(InputStream stream){}
    QC_SERIALIZABLE->addStaticMethod("deserializeToData", (q_func_n_t)static_Serializable_deserializeToData_C11InputStream, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, hashdeclSerializationInfo->getTypeInfo(false), 1, QC_INPUTSTREAM->getTypeInfo(), QORE_PARAM_NO_ARG, "stream");

    // static hash<SerializationInfo> Serializable::deserializeToData(binary bin){}
    QC_SERIALIZABLE->addStaticMethod("deserializeToData", (q_func_n_t)static_Serializable_deserializeToData_Vy, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, hashdeclSerializationInfo->getTypeInfo(false), 1, binaryTypeInfo, QORE_PARAM_NO_ARG, "bin");

    // static nothing Serializable::serialize(auto val, OutputStream stream){}
    QC_SERIALIZABLE->addStaticMethod("serialize", (q_func_n_t)static_Serializable_serialize_VAC12OutputStream, Public, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, autoTypeInfo, QORE_PARAM_NO_ARG, "val", QC_OUTPUTSTREAM->getTypeInfo(), QORE_PARAM_NO_ARG, "stream");

    // static binary Serializable::serialize(auto val){}
    QC_SERIALIZABLE->addStaticMethod("serialize", (q_func_n_t)static_Serializable_serialize_VA, Public, QCF_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 1, autoTypeInfo, QORE_PARAM_NO_ARG, "val");

    // static hash<SerializationInfo> Serializable::serializeToData(auto val){}
    QC_SERIALIZABLE->addStaticMethod("serializeToData", (q_func_n_t)static_Serializable_serializeToData_VA, Public, QCF_RET_VALUE_ONLY, QDOM_DEFAULT, hashdeclSerializationInfo->getTypeInfo(false), 1, autoTypeInfo, QORE_PARAM_NO_ARG, "val");

    return QC_SERIALIZABLE;
}
