/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file QC_NormalMethod.cpp NormalMethod class definition */
/*
    Qore Programming Language

    Copyright (C) 2003 - 2023 Qore Technologies, s.r.o.

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

    Note that the Qore library is released under a choice of three open-source
    licenses: MIT (as above), LGPL 2+, or GPL 2+; see README-LICENSE for more
    information.
*/

#include <qore_reflection.h>
#include "QC_NormalMethod.h"
#include "QC_Class.h"
/* Qore class Qore::Reflection::NormalMethod */

qore_classid_t CID_NORMALMETHOD;
QoreClass* QC_NORMALMETHOD;

// auto NormalMethod::call(object obj, ...){}
static QoreValue NormalMethod_call_VoVV(QoreObject* self, QoreReflectionMethod* m, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    QoreObject* obj = HARD_QORE_VALUE_OBJECT(args, 0);
# 135 "QC_NormalMethod.qpp"
    // get read access to program object
    QoreExternalProgramContextHelper tch(xsink, m->pgm);
    if (*xsink) {
        return QoreValue();
    }

    // we have to explictly test object compatibility here, otherwise QoreObject::evalMethod() will throw
    // a confusing exception
    if (check_call(obj, *m->m, xsink)) {
        return QoreValue();
    }

    ReferenceHolder<QoreListNode> call_args(xsink);

    // if there are arguments to pass
    if (num_args(args) > 1) {
        // create argument list by copying current list
        call_args = args->copyListFrom(1);
    }

    // the method ptr must be valid
    return obj->evalMethod(*m->m, m->m->getClass(), *call_args, xsink);
}

// auto NormalMethod::callArgs(object obj, *softlist<auto> argv){}
static QoreValue NormalMethod_callArgs_VoC15_softlist_auto_(QoreObject* self, QoreReflectionMethod* m, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    QoreObject* obj = HARD_QORE_VALUE_OBJECT(args, 0);
    const QoreListNode* argv = get_param_value(args, 1).get<const QoreListNode>();
# 176 "QC_NormalMethod.qpp"
    // get read access to program object
    QoreExternalProgramContextHelper tch(xsink, m->pgm);
    if (*xsink) {
        return QoreValue();
    }

    // we have to explictly test object compatibility here, otherwise QoreObject::evalMethod() will throw
    // a confusing exception
    if (check_call(obj, *m->m, xsink)) {
        return QoreValue();
    }

    // the method ptr must be valid
    return obj->evalMethod(*m->m, m->m->getClass(), argv, xsink);
}

// NormalMethod::constructor(Class cls, string method) {}
static void NormalMethod_constructor_C5ClassVs(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    HARD_QORE_VALUE_OBJ_DATA(cls, QoreReflectionClass, args, 0, CID_CLASS, "NormalMethod::constructor()", "Class", xsink);
    if (*xsink)
        return;
    const QoreStringNode* method = HARD_QORE_VALUE_STRING(args, 1);
# 55 "QC_NormalMethod.qpp"
ReferenceHolder<QoreReflectionClass> holder(cls, xsink);

    TempEncodingHelper nmethod(method, QCS_DEFAULT, xsink);
    if (*xsink) {
        return;
    }

    // get exclusive access to program object
    QoreExternalProgramContextHelper pch(xsink, cls->pgm);
    if (*xsink) {
        return;
    }

    // the class pointer must be valid here
    ReferenceHolder<QoreReflectionMethod> m(new QoreReflectionMethod(cls->cls, nmethod->c_str(), xsink), xsink);
    if (*xsink) {
        return;
    }

    if (m->mtype != MT_Normal) {
        xsink->raiseException("METHOD-ERROR", "method '%s()' in class '%s' has type '%s'; expecting 'normal'", nmethod->c_str(), cls->cls->getName(), m->getType());
        return;
    }

    self->setPrivate(CID_NORMALMETHOD, m.release());
}

// NormalMethod::constructor(string class_name, string method) {}
static void NormalMethod_constructor_VsVs(QoreObject* self, const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* class_name = HARD_QORE_VALUE_STRING(args, 0);
    const QoreStringNode* method = HARD_QORE_VALUE_STRING(args, 1);
# 95 "QC_NormalMethod.qpp"
TempEncodingHelper npath(class_name, QCS_DEFAULT, xsink);
    if (*xsink) {
        return;
    }

    TempEncodingHelper nmethod(method, QCS_DEFAULT, xsink);
    if (*xsink) {
        return;
    }

    ReferenceHolder<QoreReflectionMethod> m(new QoreReflectionMethod(npath->c_str(), nmethod->c_str(), xsink), xsink);
    if (*xsink) {
        return;
    }

    if (m->mtype != MT_Normal) {
        xsink->raiseException("METHOD-ERROR", "method '%s()' in class '%s' has type '%s'; expecting 'normal'", nmethod->c_str(), npath->c_str(), m->getType());
        return;
    }

    self->setPrivate(CID_NORMALMETHOD, m.release());
}

DLLLOCAL void preinitNormalMethodClass() {
    QC_NORMALMETHOD = new QoreClass("NormalMethod", "::Qore::Reflection::NormalMethod", QDOM_REFLECTION);
    CID_NORMALMETHOD = QC_NORMALMETHOD->getID();
    QC_NORMALMETHOD->setSystem();
}

DLLLOCAL QoreClass* initNormalMethodClass(QoreNamespace& ns) {
    if (!QC_NORMALMETHOD)
        preinitNormalMethodClass();

    // set parent class
    assert(QC_ABSTRACTMETHOD);
    QC_NORMALMETHOD->addBuiltinVirtualBaseClass(QC_ABSTRACTMETHOD);

    // auto NormalMethod::call(object obj, ...){}
    QC_NORMALMETHOD->addMethod("call", (q_method_n_t)NormalMethod_call_VoVV, Public, QCF_USES_EXTRA_ARGS, QDOM_DEFAULT, autoTypeInfo, 1, objectTypeInfo, QORE_PARAM_NO_ARG, "obj");

    // auto NormalMethod::callArgs(object obj, *softlist<auto> argv){}
    QC_NORMALMETHOD->addMethod("callArgs", (q_method_n_t)NormalMethod_callArgs_VoC15_softlist_auto_, Public, QCF_NO_FLAGS, QDOM_DEFAULT, autoTypeInfo, 2, objectTypeInfo, QORE_PARAM_NO_ARG, "obj", softAutoListOrNothingTypeInfo, QORE_PARAM_NO_ARG, "argv");

    // NormalMethod::constructor(Class cls, string method) {}
    QC_NORMALMETHOD->addConstructor(NormalMethod_constructor_C5ClassVs, Public, QCF_NO_FLAGS, QDOM_DEFAULT, 2, QC_CLASS->getTypeInfo(), QORE_PARAM_NO_ARG, "cls", stringTypeInfo, QORE_PARAM_NO_ARG, "method");

    // NormalMethod::constructor(string class_name, string method) {}
    QC_NORMALMETHOD->addConstructor(NormalMethod_constructor_VsVs, Public, QCF_NO_FLAGS, QDOM_DEFAULT, 2, stringTypeInfo, QORE_PARAM_NO_ARG, "class_name", stringTypeInfo, QORE_PARAM_NO_ARG, "method");

    return QC_NORMALMETHOD;
}
